/*
** Module   : MajicAccounts.java
** Abstract : Defines a custom account administration extension interface
**            for the admin server.
**
** Copyright (c) 2009-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 NVS 20090522   @42485 Created initial version with mostly stubs.
** 002 CA  20090529   @42495 Added implementation for the exports needed by 
**                           the client-side plugin.
** 003 CA  20090602   @42514 Added validation code for the extension fields. 
**                           Implemented the setExtension and addExtension 
**                           methods. Modified cloneExtension to set a new 
**                           user-code value.
** 004 CA  20090604   @42580 Modified add/set/clone extension methods to throw
**                           an exception if validation fails or a server-side
**                           maintained field can't be computed. Show a client
**                           side message if any persistence exceptions are
**                           encountered.
** 005 CA  20090609   @42614 After persistence-related methods finish work, 
**                           they must remove the session's state synchronizer.
**                           Fixed validation for top-dir field. When saving 
**                           the extensions, if the field is null, it must be 
**                           skipped or removed, if was defined - same for
**                           adding extensions.
** 006 NVS 20090807   @43547 Added getCompanyName() and setCompanyName().
** 007 CA  20090825   @43752 Fixed a bug in listRecords(...), when both record
**                           id and record description are received as filter.
**                           Added javadoc for getFilter(String, BDT).
** 008 CA  20090909   @43851 Removed file or folder validation for extensions
**                           related to files (file-viewer, file-editor, 
**                           top-dir).
** 009 NVS 20090915   @43903 Fixed user-num validation logic that prevented
**                           creation of new user records in xt-notes when
**                           creating new accounts.
**                           CA: temporary fix in checkXtNotes to allow record  
**                           saving and updating.
** 010 NVS 20090915   @43956 Fixed the transaction manager environment creation
**                           for all accesses to the database.
** 011 ECF 20090924   @44025 Fixed database access and some error handling.
**                           Calculation of a new key for the xt_notes table
**                           was based on a broken query, leading to unique
**                           constraint violations. Text values in some HQL
**                           queries needed uppercasing and right trimming.
**                           StopConditionException thrown by StandardServer
**                           invoke() had to be caught and the root cause
**                           exception exposed to enable proper error handling
**                           of checkXtNotes().
** 012 NVS 20090929   @44053 cloneExtensions() couldn't perform its duty when
**                           cloning users due to a failing check for the 
**                           target user existence. Redone.
** 013 SIY 20091116   @44332 Added helper methods for printer maintenance.
** 014 SIY 20091204   @44478 Added helper methods for function maintenance. 
** 015 SIY 20091211   @44499 Added panel menu builder. 
** 016 SIY 20091214   @44504 Replaced isAdmin() stub with real implementation.
** 017 SIY 20100419   @44820 Added refreshFunctions() and addFunction() methods.
** 018 CA  20100421   @44829 Fixed the functionDbHelperWorker - updateNotes
**                           must not be executed in batch assign mode, as it
**                           opens its own batch assign mode.
** 019 CA  20100211   @44806 Added utility to set dynamic extensions when a
**                           new account is patched to the directory.
** 020 SVL 20100211   @44581 Added some reporting functions, getRecordField(), 
**                           some utility functions for working with extensions.
** 021 CA  20100215   @44678 Fixed the getReportUserAuthForm - the user 
**                           accounts will inherit the function ACL's from its
**                           assigned groups too.
** 022 CA  20100512   @44851 Fixed the updateResponsibleDepartments function.
**                           listAllFunctions refreshes the responsible depts
**                           before retrieving the actual functions from the 
**                           DB.
** 023 SVL 20110313          Fixed name filter into listRecordsWorker. Added 
**                           the ability to add/edit/delete permit-only
**                           functions.
** 024 CA  20110411          Changes related to the "User List Report" - added
**                           include checkboxes to include only users, groups
**                           or users and groups.
** 025 SVL 20110618          Fixed getReportUserAuthForm - it wasn't showing
**                           ACLs from assigned groups if there is no ACLs
**                           assigned directly to the user.  
** 026 ECF 20130122          Migrated to new Buffer APIs.
** 027 CA  20130207          Changed to support new character implementation.
**                           Added TextOps support.
** 028 CA  20130323          DELETE statement now converts to Buffer.deleteRecord API.
** 029 OM  20130618          Migrated to new date API: date.today() instead of default constructor.
** 030 ECF 20131120          Import change.
** 031 SVL 20141205          Fixed positional parameter in getRecordFieldWorker.
** 032 SVL 20150113          Fixed more positional parameters.
** 033 SVL 20150115          Wrap pieces of code that modify data into externalProcedure for
**                           proper processing.
*/

package aero.timco.majic.admin.server;

import java.io.*;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.sql.*;
import java.util.*;

import org.hibernate.*;
import org.hibernate.type.*;

import aero.timco.majic.cfg.Config;
import aero.timco.majic.dmo.majic.*;
import aero.timco.majic.security.FunctionRights;

import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

import static com.goldencode.p2j.util.BlockManager.*;

/**
 * Defines a set of methods that enable communication between the standard
 * admin server and a custom admin server plugin, responsible for the
 * maintaining the custom account extension data in a sound state. 
 */
public class MajicAccounts 
implements AdminAccountExtension,
           MajicExtExports
{
   /** Filtering mode "greater than equal" */
   private static final int MODE_GE = 0;

   /** Filtering mode "less than equal" */
   private static final int MODE_LE = 1;

   /** Filtering mode "equal" */
   private static final int MODE_EQ = 2;

   /** instance of the class */
   private static MajicAccounts instance = null;
   
   /** synchronizer for xNotes */
   private static Object syncXtnotes = null;

   /** synchronizer for xPrint */
   private static Object syncXprinters = null;
   
   /** synchronizer for xFunc */
   private static Object syncXfunc = null;

   /** synchronizer for xMenu */
   private static Object syncXmenu = null;

   private static Object syncXimage = null;

   /**
    * Called by StandardServer at startup.
    */
   public static void initialize()
   {
      instance = new MajicAccounts();
      syncXtnotes = new Object();
      syncXprinters = new Object();
      syncXfunc = new Object();
      syncXmenu = new Object();
      syncXimage = new Object();
      
      Class<?>[] ifaces = new Class[] { MajicExtExports.class };
      RemoteObject.registerNetworkServer(ifaces, instance); 
   }
   
   /**
    * Gets a copy of the interface.
    * 
    * @return  interface. 
    */
   public static AdminAccountExtension getInterface()
   {
      return instance;
   }
   
   /**
    * Inspects the given account directory node and tells whether this account
    * has custom extension data associated with it.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. Typically, an editing batch would be open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    * <p>
    * This implementation simply checks the given node to see if there are
    * any children nodes. In MAJIC, account extensions are stored as children
    * of the account node.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @return   <code>true</code> if this account possesses any extension data
    */
   public boolean isExtended(DirectoryService ds, String node)
   {
      // check the node type
      String nodeClass = ds.getNodeClass(node);
      if (!nodeClass.equals("/meta/class/user") && 
          !nodeClass.equals("/meta/class/process"))
      {
         return false;
      }
      
      // check if there are children
      String[] ch = ds.enumerateNodes(node);
      if (ch == null || ch.length == 0)
      {
         // no extensions found
         return false;
      }
      
      // this is an extended account
      return true;
   }
   
   /**
    * Adds custom account extension data to the newly created account
    * directory node.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @param    ext
    *           a container with extension data as received from the admin
    *           client extension plugin
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    *           
    * @throws   AccountExtValidationException
    *           if any exception is encountered while validating fields or 
    *           computing default field values, an exception will be thrown. 
    */
   public boolean addExtension(DirectoryService ds, String node,
                               Serializable ext)
   throws AccountExtValidationException
   {   
      // no extension provided, exit
      if (ext == null)
      {
         return true;
      }

      String accountName = node.substring(node.lastIndexOf("/") + 1);
      if (!validateExtensions(accountName, ext, true))
      {
         throw new AccountExtValidationException(
                                               AdminServerImpl.peekMessage());
      }
      
      boolean rc = true;

      // add the extension fields
      for (int i = 0; rc && i < MajicExtDef.EXT_FIELDS.length; i++)
      {
         String field = MajicExtDef.EXT_FIELDS[i][0];
         String prop = MajicExtDef.EXT_FIELDS[i][1];
         String clazz = MajicExtDef.EXT_FIELDS[i][2];
         
         String extNode = node + "/" + field;
         
         try
         {
            Field f = ext.getClass().getField(prop);
            Object val = f.get(ext);
            // if field is not specified, do not add the node
            if (val == null)
            {
               continue;
            }
            
            Attribute attr = new Attribute(
                  ds.getClassNodeAttribute(clazz, "value"),
                  new Object[] { val }
                  );
            
            rc = ds.addNode(extNode, clazz, new Attribute[] { attr });
         }
         catch (Exception e)
         {
            rc = false;
            message("error adding field " + field);
            throw new AccountExtValidationException(
                                               AdminServerImpl.peekMessage());
         }
      }
      
      return rc;
   }
   
   /**
    * Changes custom account extension data associated with the existing
    * account directory node.
    * Changes may include synchronization, addition or deletion of the whole
    * set of extension data, based on the current state of the node and the
    * provided data input.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @param    ext
    *           a container with extension data as received from the admin
    *           client extension plugin
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    *           
    * @throws   AccountExtValidationException
    *           if any exception is encountered while validating fields or 
    *           computing default field values, an exception will be thrown. 
    */
   public boolean setExtension(DirectoryService ds, String node,
                               Serializable ext)
   throws AccountExtValidationException
   {
      boolean hasExtension = isExtended(ds, node);

      // no existing extensions and none provided - exit 
      if (!hasExtension && ext == null)
      {
         return true;
      }
      
      // existing extensions, none provided - delete 
      if (hasExtension && ext == null)
      {
         return deleteExtension(ds, node);
      }
      
      // no existing extensions, new ones are provided - add 
      if (!hasExtension && ext != null)
      {
         return addExtension(ds, node, ext);
      }
      
      // existing extensions, provided new values - set
      
      String account = node.substring(node.lastIndexOf("/")  +1);
      if (!validateExtensions(account, ext, false))
      {
         throw new AccountExtValidationException(
                                               AdminServerImpl.peekMessage());
      }

      boolean rc = true;

      // set the extension fields

      // check the existing nodes
      String[] nodes = ds.enumerateNodes(node);
      Set<String> existing = new HashSet<String>(Arrays.asList(nodes));
      
      // add the extension fields
      for (int i = 0; rc && i < MajicExtDef.EXT_FIELDS.length; i++)
      {
         String field = MajicExtDef.EXT_FIELDS[i][0];
         String prop = MajicExtDef.EXT_FIELDS[i][1];
         String clazz = MajicExtDef.EXT_FIELDS[i][2];
         
         // skip the user-code and field (is set only once)
         if (field.equals("user-code"))
         {
            continue;
         }
         
         String extNode = node + "/" + field;
         
         try
         {
            Field f = ext.getClass().getField(prop);
            Object val = f.get(ext);
            
            if (val == null)
            {
               // remove the attribute, if it was set ...
               if (existing.contains(field))
               {
                  rc = ds.deleteNode(extNode);
               }
               continue;
            }

            Attribute attr = new Attribute(
                  ds.getClassNodeAttribute(clazz, "value"),
                  new Object[] { val }
                  );
            
            if (existing.contains(field))
            {
               rc = ds.setNodeAttributes(extNode, new Attribute[] { attr });
            }
            else
            {
               rc = ds.addNode(extNode, clazz, new Attribute[] { attr });               
            }
         }
         catch (Exception e)
         {
            rc = false;
            message("error setting field " + field);
            throw new AccountExtValidationException(
                                               AdminServerImpl.peekMessage());
         }
      }
      
      return rc;
   }
   
   /**
    * Clones existing custom account extension data to the newly cloned 
    * account directory node.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    * <p>
    * This implementation clones all children of the source node into target
    * node without modification.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    fromNode
    *           full account directory node path and ID of the source account
    *
    * @param    toNode
    *           full account directory node path and ID of the target account
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    * 
    * @throws   AccountExtValidationException
    *           if any exception is encountered while validating fields or 
    *           computing default field values, an exception will be thrown. 
    */
   public boolean cloneExtension(DirectoryService ds, String fromNode, 
                                 String toNode)
   throws AccountExtValidationException
   {
      // enumerate children
      String[] ch = ds.enumerateNodes(fromNode);
      if (ch == null || ch.length == 0)
      {
         // no extensions found; strange but OK
         return true;
      }
      
      // this is an extended account; clone extensions
      for (int i = 0; i < ch.length; i++)
      {
         if (!ds.copyNode(fromNode + "/" + ch[i], 
                          toNode + "/" + ch[i], false))
         {
            // directory modification error
            return false;
         }
      }

      // compute a new unique value for the user-code field
      String node = toNode;
      boolean rc = true;

      // set the new user-code value
      {
         String userCode = getNextUserCode();
         if (userCode == null)
         {
            message("reached limit of user-code values (zzz)");
            throw new AccountExtValidationException(
                  AdminServerImpl.peekMessage());
         }
         
         String unode = node + "/user-code";
         String old = ds.getNodeString(unode, "value");
         
         if (old != null)
         {
            rc = ds.setNodeString(unode, "value", userCode);
         }
         else
         {
            rc = ds.addNodeString(unode, "value", userCode);
         }
         
         if (!rc)
         {
            return false;
         }
      }

      {
         // compute unique user-num
         int userNum = checkXtNotes(-1, "", true, true);
   
         String unode = node + "/user-num";
         Integer old = ds.getNodeInteger(unode, "value");
         
         if (old != null)
         {
            rc = ds.setNodeInteger(unode, "value", userNum);
         }
         else
         {
            rc = ds.addNodeInteger(unode, "value", userNum);
         }
         
         if (!rc)
         {
            return false;
         }
      }
      
      {
         // set the created-date
         DateValue dv = new DateValue();
         
         // set the date
         String dnode = node + "/created-date";
         DateValue old = ds.getNodeDate(dnode, "value");
         
         if (old != null)
         {
            rc = ds.setNodeDate(dnode, "value", dv);
         }
         else
         {
            rc = ds.addNodeDate(dnode, "value", dv);
         }
      }

      return rc;
   }
   
   /**
    * Deletes custom account extension data associated with the account
    * directory node being deleted.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    * <p>
    * This implementation removes all children of the node assuming there are
    * no grandchildren.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    */
   public boolean deleteExtension(DirectoryService ds, String node)
   {
      // enumerate children
      String[] ch = ds.enumerateNodes(node);
      if (ch == null || ch.length == 0)
      {
         // no extensions found; strange but OK
         return true;
      }
      
      // this is an extended account; remove extensions
      for (int i = 0; i < ch.length; i++)
      {
         if (!ds.deleteNode(node + "/" + ch[i]))
         {
            // directory modification error
            return false;
         }
      }
      
      return true;
   }

   /**
    * Get the given extension of the given account.
    *
    * @param   basePath
    *          The base path for the given account. Base path is different for
    *          the different account types. E.g. "/security/accounts/users".
    * @param   account
    *          Account name.
    * @param   extension
    *          Required account extension name.
    *
    * @return  given extension of the given account.
    */
   public Object getAccountExtension(String basePath,
                                     String account,
                                     String extension)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryService();

      StringBuilder extNode = new StringBuilder(basePath);
      extNode.append("/");
      extNode.append(account);
      extNode.append("/");
      extNode.append(extension);
      Attribute[] attrs = ds.getNodeAttributes(extNode.toString());

      if (attrs == null)
         return null;

      for (int i = 0; i < attrs.length; i++)
      {
         if (attrs[i].getName().equals("value"))
         {
            return attrs[i].getValue(0);
         }
      }

      return null;
   }

   /**
    * Gets the account-extension fields for the specified account.
    * 
    * @param    account
    *           The account for which extension fields should be retrieved.
    * 
    * @return   a container with account-extension fields.
    */
   public MajicExtDef getAccountExtensions(String account)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryService();
      
      // check if the specified account exists 
      String node1 = "/security/accounts/users/" + account;
      String node2 = "/security/accounts/processes/" + account;

      boolean isUser    = ds.getNodeClass(node1) != null;
      boolean isProcess = ds.getNodeClass(node2) != null;
      
      if (!isUser && !isProcess)
      {
         return null;
      }
      
      String node = isUser ? node1 : node2;
      
      if (!isExtended(ds, node)) 
      {
         return null;
      }

      MajicExtDef ext = new MajicExtDef();
      
      // get the existing nodes
      String[] nodes = ds.enumerateNodes(node);
      Set<String> existing = new HashSet<String>(Arrays.asList(nodes));
      
      // add the extension fields to the container
      for (int i = 0; i < MajicExtDef.EXT_FIELDS.length; i++)
      {
         String field = MajicExtDef.EXT_FIELDS[i][0];
         String prop = MajicExtDef.EXT_FIELDS[i][1];

         // if the field is not defined, skip it
         if (!existing.contains(field))
         {
            continue;
         }
         
         String extNode = node + "/" + field;
         
         try
         {
            Field f = ext.getClass().getField(prop);
            
            Attribute[] attrs = ds.getNodeAttributes(extNode);
            for (int j = 0; j < attrs.length; j++)
            {
               if (attrs[j].getName().equals("value"))
               {
                  Object val = attrs[j].getValue(0);

                  if (val != null)
                  {
                     f.set(ext, val);
                  }
                  
                  break;
               }
            }
         }
         catch (Exception e)
         {
            message("error getting extension field " + field);
         } 
      }
      
      // set the description
      ext.description = ext.key == 0 ? "" : 
         getRecordDescription("XtNotes", new integer(ext.key), "key", "text");

      return ext;
   }

   /**
    * Validate the given extensions.
    * 
    * @param    account
    *           The account to which extension data belongs
    * @param    ext
    *           The extension data
    * @param    add
    *           <code>true</code> if this is a new extension.
    * 
    * @return   <code>true</code> if extension fields are valid.
    */
   public boolean validateExtensions(String account, 
                                     Serializable ext, 
                                     boolean add)
   {
      if (!(ext instanceof MajicExtDef))
      {
         message("invalid extension container received: " + 
                  ext.getClass().getName() + "; expecting: " + 
                  MajicExtDef.class.getName());

         return false;
      }
      
      MajicExtDef m = (MajicExtDef) ext;
      
      // if any field is null, set it to a default value
      if (m.fileEditor == null)
      {
         m.fileEditor = "";
      }
      if (m.fileViewer == null)
      {
         m.fileViewer = "";
      }
      if (m.entryPoint == null)
      {
         m.entryPoint = "";
      }
      if (m.userHelp == null)
      {
         m.userHelp = "";
      }
      if (m.validApps == null)
      {
         m.validApps = "";
      }
      if (m.printer == null)
      {
         m.printer = "";
      }
      if (m.defaultFile == null)
      {
         m.defaultFile = "";
      }
      if (m.topDir == null)
      {
         m.topDir = "";
      }
      if (m.userCode == null)
      {
         m.userCode = "";
      }
      if (m.description == null)
      {
         m.description = "";
      }
      
      // convert backslashes to slashes, no other validation
      m.fileViewer = m.fileViewer.replaceAll("\\\\", "/");
      
      // convert backslashes to slashes, no other validation
      m.fileEditor = m.fileEditor.replaceAll("\\\\", "/");
      
      // entry-point: must be a valid XtFunction, if specified
      if (m.entryPoint != null      && 
          m.entryPoint.length() > 0 &&
          getEntryPointDescription(m.entryPoint) == null)
      {
         message("Invalid entry-point!");
         return false;
      }

      // user-help: no validation
      
      // run-before: no validation
      if (m.userHelp == null || m.userHelp.length() == 0)
      {
         // no help, flag is off
         m.runBefore = false;
      }
      
      // browse-mask: no validation
      
      // valid-apps: no validation; mandatory
      if (m.validApps == null || m.validApps.length() == 0)
      {
         message("No apps specified.");
         return false;
      }
      // this must be a list of apps separated by commas
      // TODO: I think they were input one-per-line ... recheck

      // auto-quit: no validation
      
      // user-level: no validation
      
      // printer: valid entry in XtPrint, if specified
      if (m.printer != null      && 
          m.printer.length() > 0 && 
          getPrinterType(m.printer) == null)
      {
         message("Invalid printer!");
         return false;
      }
      
      // default-file: no validation
      
      // convert backslashes to slashes, no other validation
      m.topDir = m.topDir.replaceAll("\\\\", "/");

      // if in add mode, validate the user-code value
      if (add)
      {
         // user-code: validate uniqueness; if not, compute an unique value
         Set<String> userCodes = getUserCodes(account);

         // validate uniqueness
         if (userCodes.contains(m.userCode))
         {
            // compute new user-code value; directory is locked at this time,
            // so this new value is unique
            m.userCode = getNextUserCode(userCodes);
            if (m.userCode == null)
            {
               message("reached limit of user-code values (zzz)");
               return false;
            }
         }
      }
      
      // user-num: valid empty entry in XtNotes; add it if one doesn't exist
      try
      {
         m.userNum = checkXtNotes(m.userNum, "", true, add);
      } 
      catch (AccountExtValidationException e)
      {
         message("could not generate user-num entry in xt_notes; " +
                 e.getMessage());
         return false;
      }

      // cust-num: valid entry in Customer
      if (m.custNum != 0 && getCustomerName(m.custNum) == null)
      {
         message("THIS CUSTOMER IS NOT IN THE DATABASE.");
         return false;
      }
      
      // emp-num: valid entry in Employee
      if (m.empNum != 0 && getEmployeeName(m.empNum) == null)
      {
         message("Cannot Locate Employee " + m.empNum);
         return false;
      }
      
      // key: valid entry in XtNotes, with user's description; can't be split
      //      into multiple lines; description must be of max 40 chars
      if (m.description != null && m.description.length() > 40)
      {
         message("description can be at most 40 chars long");
         return false;
      }
      
      try
      {
         m.key = checkXtNotes(m.key, m.description, false, add);
      } 
      catch (AccountExtValidationException e)
      {
         e.printStackTrace();
         message("could not generate description entry in xt_notes; " +
                 e.getMessage());
         return false;
      }
      
      if (add)
      {
         // get the current locale date
         DateValue dv = new DateValue();
         
         // set the date
         m.createdDate = dv;
      }
      
      return true;
   }

   /**
    * Get a set of existing user-codes from the directory. If specified, the
    * received account is skipped.
    * 
    * @param    skippedAccount
    *           The account to be skipped, or null if all accounts should be
    *           checked.
    * 
    * @return   a set of user-code values
    */
   private Set<String> getUserCodes(String skippedAccount)
   {
      SecurityAdmin sa = AdminServerImpl.getSecurityAdmin();
      DirectoryService ds = AdminServerImpl.getDirectoryService();

      TaggedName[] users = sa.listUsers();
      for (int i = 0; i < users.length; i++)
      {
         users[i].setTag("USER");
      }
      TaggedName[] procs = sa.listProcesses();
      for (int i = 0; i < procs.length; i++)
      {
         procs[i].setTag("PROC");
      }
      
      ArrayList<TaggedName> accounts = new ArrayList<TaggedName>();
      accounts.addAll(Arrays.asList(users));
      accounts.addAll(Arrays.asList(procs));

      Set<String> userCodes = new HashSet<String>(users.length);
      
      for (TaggedName account : accounts)
      {
         String id = account.getName();
         
         // skip the specified account
         if (id.equals(skippedAccount))
         {
            continue;
         }
         
         // get the user-code value for this account
         String node = "";
         if ("USER".equals(account.getTag()))
         {
            node = "/security/accounts/users/";
         }
         else
         {
            node = "/security/accounts/processes/";
         }

         node = node + id + "/user-code";
         String userCode = ds.getNodeString(node, "value");
         
         if (userCode != null)
         {
            userCodes.add(userCode);
         }
      }
      
      return userCodes;
   }

   /**
    * Send the specified message to the client.
    * 
    * @param    message
    *           the message
    */
   private void message(String message)
   {
      AdminServerImpl.message(new String[] { message });
   }
   
   /**
    * Update the XtNotes entry with the specified value. If none exists,
    * create one.
    * 
    * @param    key
    *           The key to be checked or 0 if no key exists.
    * @param    value
    *           the value to be set as text.
    * @param    userOrKey
    *           <code>true</code> if the key is the user-num index, otherwise
    *           it's the key to the description
    * @param    add
    *           <code>true</code> if creating a new account
    * 
    * @return   the new key value, if no entry exists; else, the old key value
    * 
    * @throws   AccountExtValidationException
    *           if any exception is encountered while creating the xt_notes
    *           entry, an exception will be thrown
    */
   @SuppressWarnings("null")
   private int checkXtNotes(int key, String value,
                            final boolean userOrKey, 
                            final boolean add)
   throws AccountExtValidationException
   {
      String avalue;
      
      if (value == null)
      {
         avalue = userOrKey ? "<user-num>" : "";
      }
      else if (value.length() == 0 && userOrKey)
      {
         avalue = "<user-num>";
      }
      else
      {
         avalue = value;
      }
      
      final integer aKey = key == 0 ? new integer() : new integer(key);
      final character aValue = new character(avalue);
      
      Object rc = null;
      
      try
      {
         rc = StandardServer.invoke(2, new Isolatable()
         {
            public Object execute()
            {
               return new Integer
               (
                   checkXtNotes(aKey, aValue, userOrKey, add)
               ); 
            }
         });
      }
      catch (StopConditionException sce)
      {
         Throwable chained = sce;
         
         // look through all causes to determine if the underlying error was
         // an account validation problem
         while (chained != null)
         {
            if (chained instanceof AccountExtValidationException)
            {
               throw (AccountExtValidationException) chained;
            }
            chained = chained.getCause();
         }
         
         throw sce;
      }
      catch (IllegalAccessException iax)
      {
         // never the case here
      }
      catch (InvocationTargetException itx)
      {
         // never the case here
      }
      catch (java.lang.InstantiationException ix)
      {
         // never the case here
      }
    
      return ((Integer)rc).intValue();
   }
   
   /**
    * Update the XtNotes entry with the specified value. If none exists,
    * create one.
    * 
    * @param    key
    *           The key to be checked or unknown if no key exists.
    * @param    value
    *           the value to be set as text.
    * @param    userOrKey
    *           <code>true</code> if the key is the user-num index, otherwise
    *           it's the key to the description
    * @param    add
    *           <code>true</code> if creating a new account
    * 
    * @return   the new key value, if no entry exists; else, the old key value
    * 
    * @throws   AccountExtValidationException
    *           if any exception is encountered while creating the xt_notes
    *           entry, an exception will be thrown
    */
   private int checkXtNotes(final integer key, final character value,
                            boolean userOrKey, boolean add)
   throws AccountExtValidationException
   {
      // any code which calls client-side exports (i.e. any code which calls
      // at some point a LogicalTerminal.client method) must be avoided; else,
      // the code will abend, as no client-side export exists

      final XtNotes.Buf xNotes = RecordBuffer.define(XtNotes.Buf.class, "majic", 
                                                     "xNotes");
      
      final integer newKey = new integer(0);
      final logical error = new logical(false);
      
      try
      {
         externalProcedure(new Block()
         {
            public void body()
            {
               repeat(BlockManager.TransactionType.FULL, "x-notes", new Block()
               {
                  public void init()
                  {
                     RecordBuffer.openScope(xNotes);
                  }

                  public void body()
                  {
                     synchronized (syncXtnotes)
                     {
                        // if a key exists, check if it points to a valid XtNotes
                        // record
                        if (!key.isUnknown())
                        {
                           ErrorManager.silentErrorEnable();
                           new FindQuery(xNotes, "xNotes.key = ?", null,
                                 "xNotes.key asc, xNotes.seq asc",
                                 new Object[] { key }).first();
                           ErrorManager.silentErrorDisable();
                        }

                        // if no key existed or no record found, create a new one
                        if (key.isUnknown() || !xNotes._available())
                        {
                           // TO-DO: if !add, the key should've existed, but was not
                           // found. create a new record with the existing key

                           // find the largest used key
                           ErrorManager.silentErrorEnable();
                           new FindQuery(xNotes, (String) null, null,
                                 "xNotes.key asc, xNotes.seq asc",
                                 LockType.NONE).last();
                           ErrorManager.silentErrorDisable();

                           newKey.assign(0);

                           if (xNotes._available())
                           {
                              newKey.assign(xNotes.getKey());
                           }

                           // increment the largest key by 1
                           newKey.assign(MathOps.plus(newKey, 1));

                           // create the new record
                           xNotes.create();

                           // assign new values
                           ErrorManager.silentErrorEnable();
                           RecordBuffer.startBatch();
                           xNotes.setKey(newKey);
                           xNotes.setSeq(new integer(1));
                           xNotes.setText(value);
                           RecordBuffer.endBatch();
                           ErrorManager.silentErrorDisable();
                           if (ErrorManager.isErrorFlag())
                           {
                              error.assign(true);
                              BlockManager.undoLeave("x-notes");
                           }

                           BlockManager.leave("x-notes");
                        }
                        else
                        {
                           // the record exists; update it if necessary
                           newKey.assign(key);

                           if (CompareOps.isNotEqual(xNotes.getText(), value)
                                 .booleanValue())
                           {
                              xNotes.setText(value);
                           }

                           BlockManager.leave("x-notes");
                        }
                     }
                  }
               });
            }
         });
      }
      finally
      {
         TransactionManager.setProcessingQuit(true);
      }
      
      if (error.booleanValue())
      {
         throw new AccountExtValidationException(
            ErrorManager.getErrorText(1).toStringMessage());
      }
      
      return newKey.intValue();
   }

   /**
    * List the employees with the specified filters, in the specified page.
    * 
    * @param    idFilter
    *           The employee id.
    * @param    nameFilter
    *           The employee name.
    * @param    page
    *           The current page
    * @param    pageSize
    *           The page size.
    * 
    * @return   a list with employee information, or <code>null</code> if none
    *           found.
    */
   public TaggedName[] listEmployees(String idFilter, 
                                     String nameFilter, 
                                     int    page,
                                     int    pageSize)
   {
      BaseDataType f1 = 
         idFilter == null || idFilter.trim().length() == 0 ? null 
               : new integer(idFilter.trim());
      BaseDataType f2 = 
         nameFilter == null || nameFilter.trim().length() == 0 ? null 
               : new character(nameFilter.trim() + "%");
      
      return listRecords("Employee", 
                         f1, 
                         f2, 
                         page, 
                         pageSize, 
                         0, 
                         MODE_GE, 
                         "employeeNumber", "name");
   }

   /**
    * List the customers with the specified filters, in the specified page.
    * 
    * @param    idFilter
    *           The customer id.
    * @param    nameFilter
    *           The customer name.
    * @param    page
    *           The current page
    * @param    pageSize
    *           The page size.
    * 
    * @return   a list with customer information, or <code>null</code> if none
    *           found.
    */
   public TaggedName[] listCustomers(String idFilter, 
                                     String nameFilter, 
                                     int    page,
                                     int    pageSize)
   {
      BaseDataType f1 = 
         idFilter == null || idFilter.trim().length() == 0 ? null 
               : new integer(idFilter.trim());
      BaseDataType f2 = 
         nameFilter == null || nameFilter.trim().length() == 0 ? null 
               : new character(nameFilter.trim() + "%");
      
      return listRecords("Customer", 
                         f1, 
                         f2, 
                         page, 
                         pageSize, 
                         0, 
                         MODE_GE, 
                         "customerNumber", "name");
   }

   /**
    * Refresh function names in function plugins.
    */
   public void refreshFunctions()
   {
      AdminServerImpl.refreshPlugin("majic-function");
   }

   /**
    * @see aero.timco.majic.admin.server.MajicExtExports#addFunction(java.lang.String)
    */
   public boolean addFunction(String name)
   {
      return addFunction(name, "");
   }

   /**
    * List the entry-points with the specified filters, in the specified page.
    * 
    * @param    idFilter
    *           The entry-point id.
    * @param    nameFilter
    *           The entry-point description.
    * @param    page
    *           The current page
    * @param    pageSize
    *           The page size.
    * 
    * @return   a list with entry-point information, or <code>null</code> if
    *           none found.
    */
   public TaggedName[] listEntryPoints(String idFilter, 
                                       String nameFilter, 
                                       int    page,
                                       int    pageSize)
   {
      BaseDataType f1 = 
         idFilter == null || idFilter.trim().length() == 0 ? null 
               : new character(idFilter.trim() + "%");
      BaseDataType f2 = 
         nameFilter == null || nameFilter.trim().length() == 0 ? null 
               : new character(nameFilter.trim() + "%");

      return listRecords("XtFunction",
                         f1, 
                         f2, 
                         page, 
                         pageSize, 
                         0, 
                         MODE_GE, 
                         "functionName", "functionDescription");
   }

   /**
    * List the printers with the specified filters, in the specified page.
    * 
    * @param   idFilter
    *          The printer id.
    * @param   nameFilter
    *          The printer description.
    * @param   page
    *          The current page
    * @param   pageSize
    *          The page size.
    * 
    * @return  a list with printer information, or <code>null</code> if none 
    *          found.
    */
   public TaggedName[] listPrinters(String idFilter, 
                                    String nameFilter, 
                                    int    page,
                                    int    pageSize)
   {
      BaseDataType f1 = 
         idFilter == null || idFilter.trim().length() == 0 ? null 
               : new character(idFilter.trim() + "%");
      BaseDataType f2 = 
         nameFilter == null || nameFilter.trim().length() == 0 ? null 
               : new character(nameFilter.trim() + "%");
      
      return listRecords("XtPrint", 
                         f1, 
                         f2, 
                         page, 
                         pageSize, 
                         0, 
                         MODE_GE, 
                         "printerName", "printerType");
   }

   /**
    * List all printers.
    * 
    * @return   array of printer records.
    */
   public TaggedName[] listAllPrinters()
   {
      return listRecords("XtPrint",  null, null, 0, -1, 0, 
                         MODE_GE,
                         "printerName",
                         "printerType",
                         "infile",
                         "outfile",
                         "startStr",
                         "endStr", 
                         "startProc", 
                         "endProc","printSrvr");
   }

   /**
    * List all functions.
    * 
    * @return   array of printer records.
    */
   public TaggedName[] listAllFunctions()
   {
      // reload the functions from the directory
      Config.refreshFunctionsAndDepartments();

      final int dataSize = 16;
      
      TaggedName[] funcs = listRecords("XtFunction",  null, null, 0, -1, 1,
                                       MODE_GE,
                                       "functionName",
                                       "' '",
                                       "functionDescription",
                                       "type",
                                       "functionId",
                                       "keyAccess",
                                       "printerName",
                                       "runProgram",
                                       "mdate",
                                       "transactionOk",
                                       "symixFunction",
                                       "priority",
                                       "runCall",
                                       "permitFunction", 
                                       "typeDescription", 
                                       "approvalCode"
                                       );

      // create the set of existing xt functions  
      List<TaggedName> funcsList = new ArrayList<TaggedName>();
      Set<String> existingFunctions = new HashSet<String>();
      for (TaggedName func : funcs)
      {
         funcsList.add(func);
         String desc = func.get(2); // function description,
                                    // like "Service Order Entry"
         existingFunctions.add(desc);
      }

      // get functions from the directory
      String[] directoryFunctions = Config.getFunctions();
      for (String func : directoryFunctions)
      {
         if (!existingFunctions.contains(func))
         {
            funcsList.add(new TaggedName(
                  "U",      // status "U" means that it is a permit-only function
                  "",
                  func,
                  "",
                  "",
                  "",
                  "",
                  "",
                  "",
                  "",
                  "",
                  "",
                  "",
                  "",
                  "",
                  ""));
         }
      }

      funcs = funcsList.toArray(new TaggedName[0]);

      // for all functions, add the responsible department info from the
      // directory
      for (int i = 0; i < funcs.length; i++)
      {
         TaggedName fun = funcs[i];
      
         String[] data = new String[dataSize + 1];
         for (int idx = 0; idx < dataSize; idx++)
         {
            data[idx] = fun.get(idx);
         }
         
         String function = fun.get(2);
         
         data[dataSize] = Config.getDeptForFunction(function);
         
         funcs[i] = new TaggedName(data);
      }
      
      return funcs;
   }

   /**
    * Get the function id into the directory (functions are stored into
    * /majic/functions/&lt;id&gt; branches)
    *
    * @param  name
    *         The name of the function which id should be returned.
    *
    * @return The id of the target function or <code>null</code> if there is
    *         no function with the given name. 
    */
   public String getFunctionDirectoryId(String name)
   {
      DirectoryService ds = DirectoryService.getInstance();
      String fnode = "/majic/functions/";
      String[] nodes = ds.enumerateNodes(fnode);

      for (String node : nodes)
      {
         StringBuilder sb = new StringBuilder(fnode);
         sb.append(node);
         String functionName = ds.getNodeString(sb.toString(), "values", 0);

         if (name.equals(functionName))
            return node;
      }

      return null;
   }
   
   /**
    * Retrieve additional components of the function.
    * 
    * @param    name
    *           Function name.
    *           
    * @return   Container with function components - notes, user and owner
    *           security, user and owner levels.
    */
   public TaggedName retrieveFunctionComponents(String name)
   {
      TaggedName result = new TaggedName("", "", "", "", "", "");
      
      TaggedName[] res1 = listRecords("XtFunction",  
                                      new character(name), 
                                      null, 
                                      0, 
                                      -1, 
                                      0,
                                      MODE_EQ,
                                      "functionName",
                                      "userSecurity",
                                      "ownSecurity",
                                      "userLevel", 
                                      "ownLevel", 
                                      "key");
      
      if (res1 != null && res1.length > 0)
      {
         result.set(0, res1[0].get(0)); // name 
         result.set(1, res1[0].get(1)); // user security
         result.set(2, res1[0].get(2)); // owner security
         result.set(3, res1[0].get(3)); // user level
         result.set(4, res1[0].get(4)); // owner level

         int num = 0;
         
         try
         {
            num = Integer.parseInt(res1[0].get(5)); // key
            
            if (num > 0)
            {
               res1 = listRecords("XtNotes",  
                                  new integer(res1[0].get(5)), null, 0, -1, 0,
                                  MODE_EQ,
                                  "key",
                                  "seq", 
                                  "text"); 
               
               if (res1 != null && res1.length > 0)
               {
                  // sort by sequence number
                  Arrays.sort(res1, new Comparator<TaggedName>()
                  {
                     @Override
                     public int compare(TaggedName o1, TaggedName o2)
                     {
                        return o1.get(1).compareTo(o2.get(1));
                     }
                  });
                  
                  StringBuilder buf = new StringBuilder();
                
                  // concatenate in one string
                  for (TaggedName res : res1)
                  {
                     buf.append(res.get(2).trim()).append('\n');
                  }
                  
                  result.set(5, buf.toString());
               }
            }
         }
         catch (Exception e)
         {
            return result;
         }
      }

      return result;
   }
   
   /**
    * List menus.
    * 
    * @param    name
    *           Optional menu name. If this parameter is <code>null</code>
    *           then all records are returned, but only menu name and menu
    *           description fields are filled.
    * 
    * @return   array of printer records.
    */
   public MenuDef[] listMenus(final String name)
   {
      MenuDef[] res = null;
      
      try
      {
         Object rc = StandardServer.invoke(2, new Isolatable()
         {
            public Object execute()
            {
               return listMenusWorker(name); 
            }
         });
         
         res = (MenuDef[]) rc;
      }
      catch (StopConditionException sce)
      {
         Throwable chained = sce;
         
         // look through all causes to determine if the underlying error was
         // an account validation problem
         while (chained != null)
         {
            if (chained instanceof AccountExtValidationException)
            {
               throw (AccountExtValidationException) chained;
            }
            chained = chained.getCause();
         }
         
         throw sce;
      }
      catch (IllegalAccessException iax)
      {
         // never the case here
      }
      catch (InvocationTargetException itx)
      {
         // never the case here
      }
      catch (java.lang.InstantiationException ix)
      {
         // never the case here
      }
    
      return (res == null) ? new MenuDef[0] : res;
   }

   /**
    * Menu records worker routine.
    * 
    * @param    name
    *           Optional menu name. If this parameter is <code>null</code>
    *           then all records are returned, but only menu name and menu
    *           description fields are filled.
    * 
    * @return   list of menu items for specified menu.
    */
   protected Object listMenusWorker(String name)
   {
      Persistence p = null;
      
      try
      {
         p = PersistenceFactory.getInstance("majic");
      }
      catch (Exception e) 
      {
         message(e.getMessage());
         message("error connecting to majic database");
         return null;
      }
      
      Object[] values = new Object[] {};
      Type[] types = new Type[] {};
      
      StringBuilder hql = new StringBuilder("from aero.timco.majic.dmo.majic.XtImage s ");

      if (name != null)
      {
         hql.append("where rtrim(s.menuName) = ?0");
         values = new Object[] {new character(name)};
         types = DBUtils.makeTypeArray(values);
      }
      
      hql.append(" order by upper(rtrim(s.menuName, ' \\t\\n\\r'))");

      MenuDef[] tn = null;

      try
      {
         ScrollableResults s = p.scroll(hql.toString(), 
                                        values, 
                                        types);
         if (s.first())
         {            
            ArrayList<MenuDef> list = new ArrayList<MenuDef>();

            do
            {
               list.add(makeMenuDef((XtImage) s.get(0), name != null));
            }
            while (s.next());

            tn = list.toArray(new MenuDef[list.size()]);
         }
      }
      catch (Exception e) 
      {
         message("error retrieving XtImage records");
      }
      finally
      {
         // remove the state synchronizer
         Session session = SessionManager.get().getSession();
         session.registerSynchronizer(null);
      }
      
      return tn;      
   }

   /**
    * Create {@link MenuDef} instance from {@link XtImage}.
    * 
    * @param    xImage
    * @param    full
    *           if <code>true</code> then all components of the
    *           {@link MenuDef} are loaded.
    * 
    * @return   Built instance.
    */
   protected MenuDef makeMenuDef(XtImage xImage, boolean full)
   {
      MenuDef def = new MenuDef();

      if (full)
      {
         copyFrom(xImage, def);
         
      }
      else
      {
         def.setName(xImage.getMenuName().toStringMessage());
         def.setDescription(xImage.getMenuDescription().toStringMessage());
         def.setNarrow(xImage.isNarrowMenu().booleanValue());
      }
      
      def.setComplete(full);
         
      return def;
   }

   /**
    * Add function with specified name and responsible departments into the
    * directory.
    *
    * @param name
    *        Function name.
    * @param respDept
    *        Responsible departments.
    *
    * @return <code>true</code> if function already exists (responsible
    *         departments will not be updated in this case) or was
    *         successfully added.
    */
   private boolean addFunction(String name, String respDept)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryServiceChecked();

      if (ds == null)
      {
         return false;
      }

      String base = "/majic/functions";
      String[] funcs = ds.enumerateNodes(base);
      if (funcs == null)
      {
         return false;
      }

      int lastId = -1;

      // read the function names
      for (String string : funcs)
      {
         StringBuilder sb = new StringBuilder(base);
         sb.append("/");
         sb.append(string);
         String func = ds.getNodeString(sb.toString(), "values", 0);

         if (!func.equals(name))
         {
            int id = Integer.parseInt(string);
            if (id > lastId)
               lastId = id;
         }
         else
         {
            // function already exists in directory, nothing to do
            return true;
         }
      }

      lastId++;

      boolean rc = ds.openBatch(base);

      if (!rc)
      {
         message("Can't lock directory branch. Try later.");
         return false;
      };

      // build node ID
      String node = new String(base + "/" + lastId);
      boolean res = false;
      try
      {
         Attribute[] attrs = new Attribute[1];

         attrs[0] = new Attribute(ds.getClassNodeAttribute("strings", "values"),
                                  new Object[] {name, respDept});

         res = ds.addNode(node, "strings", attrs);
      }
      finally
      {
         boolean closeRes = ds.closeBatch(res);
         res = res && closeRes;
      }

      return res;
   }

   /**
    * Copy data into <code>XtImage</code> instance.
    * 
    * @param    xImage
    */
   private static void copyTo(MenuDef def, XtImage xImage)
   {
      xImage.setMenuName(new character(def.getName()));
      xImage.setMenuDescription(new character(def.getDescription()));
      xImage.setNarrowMenu(new logical(def.isNarrow()));
      xImage.setLeftCount(new integer(def.getLeftCount()));
      xImage.setRightCount(new integer(def.getRightCount()));

      xImage.setEntryList(new character(def.mergeHotKeys()));
      
      for (int i = 0; i < MenuDef.VECTOR_SIZE; i++)
      {
         xImage.setEntryDescription(i, new character(def.getEntry(i)));
         xImage.setFunctionList(i, new character(def.getFunction(i)));
         xImage.setTypeList(i, new character(def.getType(i)));
         xImage.setNewDescription(i, new character(def.getNewDesc(i)));
      }
   }

   private static void copyFrom(XtImage xImage, MenuDef def)
   {
      def.setName(xImage.getMenuName().toStringMessage());
      def.setDescription(xImage.getMenuDescription().toStringMessage());
      def.setNarrow(xImage.isNarrowMenu().booleanValue());
      def.setLeftCount(xImage.getLeftCount().intValue());
      def.setRightCount(xImage.getRightCount().intValue());
      
      for (int i = 0; i < MenuDef.VECTOR_SIZE; i++)
      {
         def.setHotKey(i, TextOps.entry(i + 1, xImage.getEntryList()).toStringMessage());
         def.setEntryDesc(i, xImage.getEntryDescription(i).toStringMessage());
         def.setFuncList(i, xImage.getFunctionList(i).toStringMessage());
         def.setTypeList(i, xImage.getTypeList(i).toStringMessage());
         def.setNewDesc(i, xImage.getNewDescription(i).toStringMessage());
      }
   }
   
   /**
    * Validate function key access.
    * 
    * @param    name
    *           Function name.
    * @param    keys
    *           Function access keys.
    * @param    userLevel
    *           User security level.
    * @return   <code>true</code> if keys are valid (i.e. not used by another
    *           function with user level less than or equal to provided one).
    */
   public boolean validateKeyAccess(String name, String keys,
                                    String userLevel)
   {
      int level = 0;
      
      try
      {
         level = Integer.parseInt(userLevel);
      }
      catch (NumberFormatException e)
      {
         // ignore it
      }
      
      TaggedName[] res1 = listRecords("XtFunction",  
                                      new character(keys), 
                                      new integer(level), 
                                      0, 
                                      -1, 
                                      0,
                                      MODE_LE,
                                      "keyAccess",
                                      "userLevel", 
                                      "functionName");
      
      boolean res = true;
      
      if (res1 != null && res1.length > 0)
      {
         for (TaggedName rec : res1)
         {
            if (rec.get(2).equals(name))
               continue;
            
            res = false;
            break;
         }
      }
      
      return res;
   }

   /**
    * Get user accounts information together with the specified extension
    * fields.
    *
    * @param   extensions
    *          Extension fields that should be retrieved.
    *
    * @return  a set of {@link TaggedName} instances which contain required
    *          account infomation or <code>null</code>. 
    */
   public TaggedName[] getUsersWithExtensions(String[] extensions)
   {
      TaggedName[] users = AdminServerImpl.listUsers();
      if (users == null)
         return null;

      for (int i = 0; i < users.length; i++)
      {
         TaggedName user = users[i];
         int len = user.length();
         Object[] arr = new Object[len + extensions.length];
         for (int j = 0; j < len; j++)
         {
            arr[j] = user.get(j);
         }

         for (int j = 0; j < extensions.length; j++)
         {
            String extension = extensions[j];
            Object res = getAccountExtension("/security/accounts/users",
                                             user.getName(),
                                             extension);

            arr[len + j] = res != null ? res : "";
         }

         users[i] = new TaggedName(arr);
      }

      return users;
   }

   public String getUserNameByEmployeeNumber(int employeeNumber)
   {
      TaggedName[] users = AdminServerImpl.listUsers();
      if (users == null)
         return null;

      for (TaggedName user : users)
      {
         String userName = user.getName();
         Object res = getAccountExtension("/security/accounts/users",
                                          userName,
                                          "emp-num");
         if (((Integer) employeeNumber).equals(res))
         {
            return userName;
         }
      }

      return null;
   }

   /**
    * Get the "User Authorization By Function" report with the given
    * parameters.
    *
    * @param startUser
    *        Starting User ID.
    * @param endUser
    *        Ending User ID.
    * @param startEmpNum
    *        Starting Employee Number.
    * @param endEmpNum
    *        Ending Employee Number.
    * @param startSkill
    *        Starting Skill.
    * @param endSkill
    *        Ending Skill.
    * @param showTerminated
    *        Whether terminated users should be included in the report.
    * @param include
    *        A parameter indicating which accounts to include:
    *        <ul>
    *          <li>0 for ALL users and groups</li>
    *          <li>1 for user accounts</li>
    *          <li>2 for group accounts</li>
    *        </ul>
    * 
    * @return an array of {@link TaggedName} instances, with the report rows.
    */
   public TaggedName[] getReportUserList(String startUser,
                                         String endUser,
                                         int startEmpNum,
                                         int endEmpNum,
                                         String startSkill,
                                         String endSkill,
                                         boolean showTerminated,
                                         int include)
   {
      TaggedName[] sourceList = include == 2 
                                  ? null
                                  : getUsersWithExtensions(
                                       new String[]{"emp-num", "key"});

      // if none found, reset the user list
      if (sourceList == null)
         sourceList = new TaggedName[0];

      final List<TaggedName> targetList = new ArrayList<TaggedName>();

      for (TaggedName user : sourceList)
      {
         String username = user.getName();
                                      
         if (!intoStringRange(startUser, endUser, username))
            continue;

         int empnum = 0;
         try
         {
            empnum = Integer.parseInt(user.get(2));
         }
         catch (NumberFormatException e)
         {
            // suppress;
         }

         if (!intoIntRange(startEmpNum, endEmpNum, empnum))
            continue;

         int key = 0;
         try
         {
            key = Integer.parseInt(user.get(3));
         }
         catch (NumberFormatException e)
         {
            // suppress;
         }

         character skill = null;
         date terminationDate = null;
         if (empnum > 0)
         {
            skill = (character) getRecordField("Employee",
                                               new integer(empnum),
                                               "employeeNumber",
                                               "skillCode");

            String skillStr = skill == null || skill.isUnknown() ?
                              null :
                              skill.toJavaType();
            if (!intoStringRange(startSkill, endSkill, skillStr))
               continue;

            terminationDate = (date) getRecordField("Employee",
                                                    new integer(empnum),
                                                    "employeeNumber",
                                                    "terminationDate");

            if (terminationDate != null &&
                !terminationDate.isUnknown() &&
                !showTerminated)
               continue;
         }
         else
         {
            if (!intoStringRange(startSkill, endSkill, null))
               continue;
         }

         character description = null;
         if (key > 0)
         {
            description = (character) getRecordField("XtNotes",
                                                     new integer(key),
                                                     "key",
                                                     "text");
         }

         String[] arr =
         {
            username,
            user.get(1), //name
            Integer.toString(empnum),
            TaggedName.prettyPrintNoUnknows(skill),
            TaggedName.prettyPrintNoUnknows(terminationDate),
            TaggedName.prettyPrintNoUnknows(description)
         };

         TaggedName userinfo = new TaggedName(arr);
         targetList.add(userinfo);
      }

      sourceList = include == 1 ? null : AdminServerImpl.listGroups();

      // if none found, reset the group list
      if (sourceList == null)
         sourceList = new TaggedName[0];
      
      for (TaggedName group : sourceList)
      {
         String groupname = group.getName();

         if (!intoStringRange(startUser, endUser, groupname))
            continue;

         String[] arr =
         {
            groupname,
            "",
            "0",
            "",
            "",
            group.getTag() // description
         };

         TaggedName groupinfo = new TaggedName(arr);
         targetList.add(groupinfo);
      }

      //sort by username
      Collections.sort(targetList);

      return targetList.toArray(new TaggedName[0]);
   }

   /**
    * Get the Majic version.
    *
    * @return See above.
    */
   public String getMajicVersion()
   {
      return getRecordDescription("LastTransaction",
                                  new integer(3),
                                  "lastTransactionKey",
                                  "transactionFile");
   }

   /**
    * Get the common report header.
    *
    * @return See above.
    */
   public String getCommonReportHeader()
   {
      return getRecordDescription("Parameters",
                                  new integer(0),
                                  "parameterKey",
                                  "reportHeader");
   }

   /**
    * Add or update printer definition.
    * 
    * @param    tn
    *           Printer definition.
    * @param    mode
    *           Mode of operation: {@link #MODE_ADD} - add record,
    *           {@link #MODE_DELETE} - remove record, {@link #MODE_EDIT} -
    *           update record.
    * 
    * @return   error code: 0 - means success, 1 - printer is already present,
    *           other values mean some other kinds of DB issues.
    */
   public int printerDbHelper(final TaggedName tn, final int mode)
   {
      Integer rc = null;
      
      try
      {
         rc = (Integer) StandardServer.invoke(2, new Isolatable()
         {
            public Object execute()
            {
               return new Integer(printerDbHelperWorker(tn, mode)); 
            }
         });
      }
      catch (StopConditionException sce)
      {
         Throwable chained = sce;
         
         // look through all causes to determine if the underlying error was
         // an account validation problem
         while (chained != null)
         {
            if (chained instanceof AccountExtValidationException)
            {
               throw (AccountExtValidationException) chained;
            }
            chained = chained.getCause();
         }
         
         throw sce;
      }
      catch (IllegalAccessException iax)
      {
         // never the case here
      }
      catch (InvocationTargetException itx)
      {
         // never the case here
      }
      catch (java.lang.InstantiationException ix)
      {
         // never the case here
      }
    
      if (rc == null)
         return RC_CANT_UPDATE;
      
      return rc.intValue();
   }
   
   /**
    * Add, delete or update printer records worker routine.
    * 
    * @param    tn
    *           Printer definition.
    * @param    mode
    *           Mode of operation: {@link #MODE_ADD} - add record,
    *           {@link #MODE_DELETE} - remove record, {@link #MODE_EDIT} -
    *           update record.
    *           
    * @return   error code: 0 - means success, 1 - printer is already present,
    *           other values mean some other kinds of DB issues.
    */
   private int printerDbHelperWorker(final TaggedName tn, final int mode)
   {
      final XtPrint.Buf xPrint = RecordBuffer.define(XtPrint.Buf.class, "majic",
                                                     "xPrint");
      final int[] rc = {RC_OK};

      try
      {
         externalProcedure(new Block()
         {
            public void body()
            {
               repeat(BlockManager.TransactionType.FULL, "x-print", new Block()
               {
                  public void init()
                  {
                     RecordBuffer.openScope(xPrint);
                  }

                  public void body()
                  {
                     synchronized (syncXprinters)
                     {
                        ErrorManager.silentErrorEnable();
                        new FindQuery(xPrint, "rtrim(xPrint.printerName) = ?", null,
                              "xPrint.printerName asc",
                              new Object[] { new character(tn.getName()) }).first();
                        ErrorManager.silentErrorDisable();

                        if (xPrint._available())
                        {
                           if (mode == MajicExtExports.MODE_ADD)
                           {
                              rc[0] = RC_ALREADY_EXISTS; // printer already exists
                              BlockManager.undoLeave("x-print");
                              return;
                           }

                           if (mode == MajicExtExports.MODE_DELETE)
                           {
                              xPrint.deleteRecord();
                              BlockManager.leave("x-print");
                              return;
                           }

                           if (mode == MajicExtExports.MODE_EDIT)
                           {
                              tagged2xPrint(tn, xPrint);
                              BlockManager.leave("x-print");
                              return;
                           }
                        }
                        else
                        {
                           if (mode == MajicExtExports.MODE_DELETE ||
                                 mode == MajicExtExports.MODE_EDIT)
                           {
                              rc[0] = RC_DOES_NOT_EXISTS;
                              BlockManager.undoLeave("x-print");
                              return;
                           }

                           if (mode == MajicExtExports.MODE_ADD)
                           {
                              xPrint.create();

                              ErrorManager.silentErrorEnable();
                              RecordBuffer.startBatch();

                              tagged2xPrint(tn, xPrint);

                              RecordBuffer.endBatch();
                              ErrorManager.silentErrorDisable();

                              if (ErrorManager.isErrorFlag())
                              {
                                 rc[0] = RC_CANT_ADD;
                                 BlockManager.undoLeave("x-print");
                              }
                              else
                                 BlockManager.leave("x-print");
                              return;
                           }
                        }
                     }
                  }
               });
            }
         });
      }
      finally
      {
         TransactionManager.setProcessingQuit(true);
      }

      return rc[0];
   }

   /**
    * Add or update function definition.
    * 
    * @param    tn
    *           Function definition.
    * @param    mode
    *           Mode of operation: {@link #MODE_ADD} - add record,
    *           {@link #MODE_DELETE} - remove record, {@link #MODE_EDIT} -
    *           update record.
    * 
    * @return   error code: 0 - means success, 1 - printer is already present,
    *           other values mean some other kinds of DB issues.
    */
   public int functionDbHelper(final TaggedName tn, final int mode)
   {
      Integer rc = null;
      
      try
      {
         rc = (Integer) StandardServer.invoke(2, new Isolatable()
         {
            public Object execute()
            {
               return new Integer(functionDbHelperWorker(tn, mode)); 
            }
         });
      }
      catch (StopConditionException sce)
      {
         Throwable chained = sce;
         
         // look through all causes to determine if the underlying error was
         // an account validation problem
         while (chained != null)
         {
            if (chained instanceof AccountExtValidationException)
            {
               throw (AccountExtValidationException) chained;
            }
            chained = chained.getCause();
         }
         
         throw sce;
      }
      catch (IllegalAccessException iax)
      {
         // never the case here
      }
      catch (InvocationTargetException itx)
      {
         // never the case here
      }
      catch (java.lang.InstantiationException ix)
      {
         // never the case here
      }
    
      if (rc == null)
         return RC_CANT_UPDATE;
      
      return rc.intValue();
   }
   
   /**
    * Add, delete or update function records worker routine.
    * 
    * @param    tn
    *           Function definition.
    * @param    mode
    *           Mode of operation: {@link #MODE_ADD} - add record,
    *           {@link #MODE_DELETE} - remove record, {@link #MODE_EDIT} -
    *           update record.
    *           
    * @return   error code: 0 - means success, 1 - printer is already present,
    *           other values mean some other kinds of DB issues.
    */
   private int functionDbHelperWorker(final TaggedName tn, final int mode)
   {
      final XtFunction.Buf xFunc  = RecordBuffer.define(XtFunction.Buf.class,
                                                        "majic", 
                                                        "xFunc");
      final XtNotes.Buf xNotes = RecordBuffer.define(XtNotes.Buf.class, "majic", 
                                                     "xNotes");
      final String function = tn.get(1).trim();
      final String respDep = tn.get(15).trim();

      final int[] rc = {RC_OK};
      
      try
      {
         externalProcedure(new Block()
         {
            public void body()
            {
               repeat(BlockManager.TransactionType.FULL, "x-func", new Block()
               {
                  public void init()
                  {
                     RecordBuffer.openScope(xFunc, xNotes);
                  }

                  public void body()
                  {
                     synchronized (syncXfunc)
                     {
                        // check whether we are handing a permit function
                        boolean isPermitFunction;

                        if (mode == MajicExtExports.MODE_ADD)
                        {
                           isPermitFunction = "permit".equals(tn.get(21));
                        }
                        else
                        {
                           isPermitFunction = tn.get(0).trim().isEmpty(); // xt_function name is empty
                        }

                        if (isPermitFunction)
                        {
                           if (mode == MajicExtExports.MODE_ADD)
                           {
                              if (getFunctionDirectoryId(function) != null)
                              {
                                 rc[0] = RC_ALREADY_EXISTS;
                                 BlockManager.undoLeave("x-func");
                              }

                              ErrorManager.silentErrorEnable();
                              new FindQuery(xFunc, "ltrim(rtrim(xFunc.functionDescription)) = ?", null,
                                    "xFunc.functionDescription asc",
                                    new Object[] { new character(function) }).first();
                              ErrorManager.silentErrorDisable();

                              if (xFunc._available())
                              {
                                 rc[0] = RC_ALREADY_EXISTS;
                                 BlockManager.undoLeave("x-func");
                              }

                              if (!updateResponsibleDepartments(function,
                                    respDep))
                              {
                                 rc[0] = RC_CANT_ADD;
                                 BlockManager.undoLeave("x-func");
                              }

                              BlockManager.leave("x-func");
                           }

                           String functionOldName = tn.get(22).trim();

                           if (mode == MajicExtExports.MODE_EDIT)
                           {
                              if (functionOldName.isEmpty())
                              {
                                 // no function name specified
                                 rc[0] = RC_CANT_UPDATE;
                                 BlockManager.undoLeave("x-func");
                              }

                              String id = getFunctionDirectoryId(functionOldName);
                              if (id == null)
                              {
                                 // function does not exist
                                 rc[0] = RC_CANT_UPDATE;
                                 BlockManager.undoLeave("x-func");
                              }

                              if (!functionOldName.equals(function))
                              {
                                 if (hasRelatedACLs(functionOldName))
                                 {
                                    // function is involved into ACL - cannot change function name
                                    rc[0] = RC_CANT_UPDATE;
                                    BlockManager.undoLeave("x-func");
                                 }
                              }

                              if (!updateFunction(id, function, respDep))
                              {
                                 // update failed
                                 rc[0] = RC_CANT_UPDATE;
                                 BlockManager.undoLeave("x-func");
                              }

                              BlockManager.leave("x-func");
                           }

                           if (mode == MajicExtExports.MODE_DELETE)
                           {
                              String id = getFunctionDirectoryId(functionOldName);
                              if (id == null)
                              {
                                 BlockManager.leave("x-func");
                              }

                              if (hasRelatedACLs(functionOldName))
                              {
                                 // function involved into ACL
                                 rc[0] = RC_CANT_DELETE;
                                 BlockManager.undoLeave("x-func");
                              }

                              if (!deleteFunction(id))
                              {
                                 // deletion failed
                                 rc[0] = RC_CANT_DELETE;
                                 BlockManager.undoLeave("x-func");
                              }

                              BlockManager.leave("x-func");
                           }
                        }
                        else
                        {
                           // regular function
                           ErrorManager.silentErrorEnable();
                           new FindQuery(xFunc, "ltrim(rtrim(xFunc.functionName)) = ?", null,
                                 "xFunc.functionName asc",
                                 new Object[] { new character(tn.getName()) }).first();
                           ErrorManager.silentErrorDisable();

                           if (xFunc._available())
                           {
                              if (mode == MajicExtExports.MODE_ADD)
                              {
                                 rc[0] = RC_ALREADY_EXISTS;
                                 BlockManager.undoLeave("x-func");
                                 return;
                              }

                              if (mode == MajicExtExports.MODE_DELETE)
                              {
                                 xFunc.deleteRecord();

                                 // the /majic/function directory entry is left intact

                                 BlockManager.leave("x-func");
                                 return;
                              }

                              if (mode == MajicExtExports.MODE_EDIT)
                              {
                                 updateNotes(xFunc, xNotes, tagged2xFunc(tn, xFunc));

                                 // the /majic/function directory entry is updated or
                                 // added
                                 if (!updateResponsibleDepartments(function, respDep))
                                 {
                                    rc[0] = RC_ALREADY_EXISTS;
                                    BlockManager.undoLeave("x-func");
                                    return;
                                 }

                                 BlockManager.leave("x-func");
                                 return;
                              }
                           }
                           else
                           {
                              if (mode == MajicExtExports.MODE_DELETE)
                              {
                                 BlockManager.undoLeave("x-func");
                                 return;
                              }

                              if (mode == MajicExtExports.MODE_EDIT ||
                                    mode == MajicExtExports.MODE_ADD)
                              {
                                 xFunc.create();

                                 ErrorManager.silentErrorEnable();
                                 RecordBuffer.startBatch();

                                 character[] notes = tagged2xFunc(tn, xFunc);

                                 RecordBuffer.endBatch();
                                 ErrorManager.silentErrorDisable();

                                 updateNotes(xFunc, xNotes, notes);

                                 if (ErrorManager.isErrorFlag())
                                 {
                                    rc[0] = RC_CANT_ADD;
                                    BlockManager.undoLeave("x-func");
                                 }
                                 else
                                 {
                                    // the /majic/function directory entry is updated
                                    // or added
                                    if (!updateResponsibleDepartments(function,
                                          respDep))
                                    {
                                       rc[0] = RC_ALREADY_EXISTS;
                                       BlockManager.undoLeave("x-func");
                                       return;
                                    }

                                    BlockManager.leave("x-func");
                                 }
                                 return;
                              }
                           }
                        }
                     }
                  }
               });
            }
         });
      }
      finally
      {
         TransactionManager.setProcessingQuit(true);
      }

      return rc[0];
   }

   /**
    * Checks whether a function has related ACL(s).
    *
    * @param  function
    *         Name of the function to be checked.
    *
    * @return <code>true</code> if the given function has related ACL(s). 
    */
   private boolean hasRelatedACLs(String function)
   {
      String[] res = AdminServerImpl.listAclInstances("majic-function");
      String name = AdminServerImpl.getServerName();

      if (name == null)
         name = "";

      for (String instance : res)
      {
         if (instance.length() == 0 || instance.equals(name))
         {
            AclDef def = AdminServerImpl.getAcl("majic-function", instance);

            for (AclDef.Acl oneAcl: def.acls)
            {
               if (oneAcl.name.equals(function))
                  return true;
            }
         }
      }

      return false;
   }

   /**
    * Update or add the "responsible departments" info for the given function.
    * 
    * @param   function
    *          Function description.
    * @param   respDept
    *          The responsible departments.
    *          
    * @return  <code>true</code> if the update succeded
    */
   private boolean updateResponsibleDepartments(String function,
                                                String respDept)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryService();
      String fnode = "/majic/functions/";
      
      String[] nodes = ds.enumerateNodes(fnode);
      String node = null;
      
      for (int i = 0; i < nodes.length; i++)
      {
         String cnode = fnode + nodes[i];
         
         String[] attrs = ds.getNodeStrings(cnode, "values");
         
         if (function.equals(attrs[0].trim()))
         {
            node = cnode;
            break;
         }
      }
      
      boolean res = true;

      if (node != null)
      {
         String[] attrs = ds.getNodeStrings(node, "values");
         // if the value was changed, update it
         if (!respDept.equals(attrs[1]))
         {
            attrs[1] = respDept;
            res = ds.openBatch(fnode);
            if (!res)
            {
               message("Can't lock directory branch. Try later."); 
               return false;
            }
            
            try
            {
               res = ds.setNodeStrings(node, "values", attrs);
            }
            finally
            {
               boolean closeRes = ds.closeBatch(res);
               res = res && closeRes;
            }   
         }
      }
      else
      {
         res = addFunction(function, respDept);
      }

      return res;
   }

   /**
    * Updates the function with the given id into the directory.
    *  
    * @param  id
    *         Id of the target function.
    * @param  function
    *         New function name.
    * @param  respDept
    *         New function responsible departments.
    *
    * @return <code>true</code> if the update was successful.
    */
   private boolean updateFunction(String id,
                                  String function,
                                  String respDept)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryService();
      String fnode = "/majic/functions/";
      StringBuilder sb = new StringBuilder(fnode);
      sb.append(id);
      String node = sb.toString();

      String[] attrs = ds.getNodeStrings(node, "values");

      if (attrs == null)
         return false;

      attrs[0] = function;
      attrs[1] = respDept;

      boolean res = ds.openBatch(fnode);
      if (!res)
      {
         message("Can't lock directory branch. Try later.");
         return false;
      }

      try
      {
         res = ds.setNodeStrings(node, "values", attrs);
      }
      finally
      {
         boolean closeRes = ds.closeBatch(res);
         res = res && closeRes;
      }

      return res;
   }

   /**
    * Deletes the function with the given id from the directory.
    *
    * @param  id
    *         Id of the target function.
    *
    * @return <code>true</code> if the function was successfully deleted.
    */
   private boolean deleteFunction(String id)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryService();
      String fnode = "/majic/functions/";
      StringBuilder sb = new StringBuilder(fnode);
      sb.append(id);

      boolean res = ds.openBatch(fnode);
      if (!res)
      {
         message("Can't lock directory branch. Try later.");
         return false;
      }

      try
      {
         res = ds.deleteNode(sb.toString());
      }
      finally
      {
         boolean closeRes = ds.closeBatch(res);
         res = res && closeRes;
      }

      return res;
   }
   
   /**
    * Update notes records.
    * 
    * @param    xFunc
    *           XtFunction record buffer.
    * @param    xNotes
    *           Notes record buffer.
    * @param    notes
    *           Notes for this record.
    */
   private void updateNotes(XtFunction.Buf xFunc,
                            XtNotes.Buf xNotes,
                            character[] notes)
   {
      integer key = xFunc.getKey();
      
      if (key.isUnknown() || key.equals(new integer(0)))
      {
         if  (notes == null || notes.length == 0)
         {
            // no existing notes and we not going to add new ones, just return
            return;
         }
         // generate new key
         ErrorManager.silentErrorEnable();
         new FindQuery(xNotes, null, null, 
                       "xNotes.key desc", 
                       new Object[0]).first();
         ErrorManager.silentErrorDisable();
         
         if (xNotes._available())
         {
            key = new integer(xNotes.getKey().getValue() + 1); 
         }
         else
         {
            key = new integer(1);
         }
      }
      else
      {
         // remove existing notes
         ErrorManager.silentErrorEnable();
         FindQuery query = new FindQuery(xNotes, "xNotes.key = ?", null, 
                                         "xNotes.key asc", 
                                         new Object[] {key});
         
         query.first();
         
         while (xNotes._available())
         {
            xNotes.deleteRecord();
            query.next();
         }
         
         ErrorManager.silentErrorDisable();
      }
      
      xFunc.setKey(key);
      
      // add notes
      for (int i = 0; i < notes.length; i++)
      {
         xNotes.create();
         
         ErrorManager.silentErrorEnable();
         RecordBuffer.startBatch();

         xNotes.setKey(key);
         xNotes.setSeq(new integer(i + 1));
         xNotes.setText(notes[i]);

         RecordBuffer.endBatch();
         ErrorManager.silentErrorDisable();
      }
      // do not commit here
   }
   
   /**
    * Parse function notes.
    * 
    * @param    string
    * 
    * @return   notes split into array of strings which contain at most 40
    *           chars.
    */
   private static character[] parseNotes(String string)
   {
      if (string == null || string.trim().equals(""))
         return null;
      
      String[] parsed = string.split("\n");
      ArrayList<character> res = new ArrayList<character>();
      
      for (int i = 0; i < parsed.length; i++)
      {
         String s = parsed[i];
         
         while (s.length() > 40)
         {
            res.add(new character(s.substring(0, 40)));
            
            s = s.substring(40);
         }
         
         res.add(new character(s));
      }
      
      return res.toArray(new character[res.size()]);
   }

   /**
    * Add or update menu definition.
    * 
    * @param    def
    *           Menu definition.
    * @param    mode
    *           Mode of operation: {@link #MODE_ADD} - add record,
    *           {@link #MODE_DELETE} - remove record, {@link #MODE_EDIT} -
    *           update record.
    * 
    * @return   error code: 0 - means success, 1 - printer is already present,
    *           other values mean some other kinds of DB issues.
    *           
    * @throws   AccountExtValidationException 
    */
   public int menuDbHelper(final MenuDef def, final int mode)
   {
      Integer rc = null;
      final String[] message = new String[] {""};
      
      try
      {
         rc = (Integer) StandardServer.invoke(2, new Isolatable()
         {
            public Object execute()
            {
               return new Integer(menuDbHelperWorker(def, mode, message)); 
            }
         });
      }
      catch (StopConditionException sce)
      {
         Throwable chained = sce;
         
         // look through all causes to determine if the underlying error was
         // an account validation problem
         while (chained != null)
         {
            if (chained instanceof AccountExtValidationException)
            {
               throw (AccountExtValidationException) chained;
            }
            chained = chained.getCause();
         }
         
         throw sce;
      }
      catch (IllegalAccessException iax)
      {
         // never the case here
      }
      catch (InvocationTargetException itx)
      {
         // never the case here
      }
      catch (java.lang.InstantiationException ix)
      {
         // never the case here
      }
    
      if (rc == null)
         return RC_CANT_UPDATE;
      
      if (rc == RC_VALIDATION_ERROR)
         throw new AccountExtValidationException(message[0]);
      
      return rc.intValue();
   }
   
   /**
    * Add, delete or update menu records worker routine.
    * 
    * @param    def
    *           Menu definition.
    * @param    mode
    *           Mode of operation: {@link #MODE_ADD} - add record,
    *           {@link #MODE_DELETE} - remove record, {@link #MODE_EDIT} -
    *           update record.
    * @param    message
    *           Storage for validation error message.
    *           
    * @return   error code: 0 - means success, 1 - printer is already present,
    *           other values mean some other kinds of DB issues.
    */
   private int menuDbHelperWorker(final MenuDef def, 
                                  final int mode, 
                                  final String[] message)
   {
      final XtImage.Buf xImage =
         RecordBuffer.define(XtImage.Buf.class, "majic", "xImage");
      final XtFunction.Buf xFunction =
         RecordBuffer.define(XtFunction.Buf.class, "majic", "xFunction");
      
      final int[] rc = {RC_OK};
      
      try
      {
         externalProcedure(new Block()
         {
            public void body()
            {
               repeat(BlockManager.TransactionType.FULL, "x-menu", new Block()
               {
                  public void init()
                  {
                     RecordBuffer.openScope(xImage);
                  }

                  public void body()
                  {
                     synchronized (syncXmenu)
                     {
                        ErrorManager.silentErrorEnable();
                        new FindQuery(xImage,
                              "rtrim(xImage.menuName) = ?",
                              null,
                              "xImage.menuName asc",
                              new Object[] { new character(def.getName()) }).first();
                        ErrorManager.silentErrorDisable();

                        if (xImage._available())
                        {
                           if (mode == MajicExtExports.MODE_ADD)
                           {
                              rc[0] = RC_ALREADY_EXISTS; // printer already exists
                              BlockManager.undoLeave("x-menu");
                              return;
                           }

                           if (mode == MajicExtExports.MODE_DELETE)
                           {
                              ErrorManager.silentErrorEnable();
                              new FindQuery(xFunction,
                                    "rtrim(xFunction.functionName) = ?",
                                    null,
                                    "xFunction.functionName asc",
                                    new Object[] { xImage.getMenuName() }).first();
                              ErrorManager.silentErrorDisable();

                              if (xFunction._available())
                              {
                                 if (xFunction.isPermitFunction().booleanValue() && !isAdmin())
                                 {
                                    rc[0] = RC_VALIDATION_ERROR;
                                    message[0] = "only administrator can delete permanent menu.";
                                    BlockManager.undoLeave("x-menu");
                                    return;
                                 }
                              }

                              xImage.deleteRecord();

                              if (xFunction._available())
                              {
                                 xFunction.deleteRecord();
                              }

                              BlockManager.leave("x-menu");
                              return;
                           }

                           if (mode == MajicExtExports.MODE_EDIT)
                           {
                              if (!validateAddOrUpdateMenu(def, message, xFunction, xImage))
                              {
                                 rc[0] = RC_VALIDATION_ERROR;
                                 BlockManager.undoLeave("x-menu");
                                 return;
                              }

                              copyTo(def, xImage);
                              BlockManager.leave("x-menu");
                              return;
                           }
                        }
                        else
                        {
                           if (mode == MajicExtExports.MODE_DELETE ||
                                 mode == MajicExtExports.MODE_EDIT)
                           {
                              rc[0] = RC_DOES_NOT_EXISTS;
                              BlockManager.undoLeave("x-menu");
                              return;
                           }

                           if (mode == MajicExtExports.MODE_ADD)
                           {
                              if (!validateAddOrUpdateMenu(def, message, xFunction, xImage))
                              {
                                 rc[0] = RC_VALIDATION_ERROR;
                                 BlockManager.undoLeave("x-menu");
                                 return;
                              }

                              xImage.create();

                              ErrorManager.silentErrorEnable();
                              RecordBuffer.startBatch();

                              copyTo(def, xImage);

                              RecordBuffer.endBatch();
                              ErrorManager.silentErrorDisable();

                              if (ErrorManager.isErrorFlag())
                              {
                                 rc[0] = RC_CANT_ADD;
                                 BlockManager.undoLeave("x-menu");
                              }
                              else
                                 BlockManager.leave("x-menu");
                              return;
                           }
                        }
                     }
                  }
               });
            }
         });
      }
      finally
      {
         TransactionManager.setProcessingQuit(true);
      }

      return rc[0];
   }
   
   /**
    * Check if current user has administrator rights.
    * 
    * @return   <code>true</code> if user is administrator.
    */
   protected boolean isAdmin()
   {
      return AdminServerImpl.getAdminDef() != null;
   }

   /**
    * Validate adding or update of the menu.
    * 
    * @param    def
    *           Menu definition.
    * @param    message
    *           Storage for validation message.
    * @param    xFunction
    *           {@link XtFunction} record buffer.
    * @param    xImage
    *           {@link XtImage} record buffer.
    *           
    * @return   <code>true</code> if record can be added or updated.
    */
   protected boolean validateAddOrUpdateMenu(MenuDef def, 
                                             String[] message, 
                                             XtFunction.Buf xFunction,
                                             XtImage.Buf xImage)
   {
      def.cleanup();
      
      Set<String> keySet = new HashSet<String>();
      
      for (String key : def.getHotKeys())
      {
         if (key.length() == 0)
            continue;
         
         if (keySet.contains(key))
         {
            message[0] = "key '" + key + "' is already assigned.";
            return false;
         }
            
         keySet.add(key);
      }
      
      ArrayList<String> checkList = new ArrayList<String>();
      
      // check functions for existence
      if (xImage._available())
      {
         for (int i = 0; i < MenuDef.VECTOR_SIZE; i++)
         {
            String function = def.getFunction(i).trim();
            
            if (function.length() > 0 && !function.equalsIgnoreCase("text"))
            {
               if (!CompareOps._isEqual(function, xImage.getFunctionList(i)))
                  checkList.add(function);
            }
         }
      }
      else
      {
         for (int i = 0; i < MenuDef.VECTOR_SIZE; i++)
         {
            String function = def.getFunction(i).trim();
            
            if (function.length() > 0 && !function.equalsIgnoreCase("text"))
               checkList.add(function);
         }
      }
      
      for (String function : checkList)
      {
         ErrorManager.silentErrorEnable();
         new FindQuery(xFunction, 
                       "rtrim(xFunction.functionName) = ?", 
                       null, 
                       "xFunction.functionName asc", 
                       new Object[] { new character(function)}).first();
         ErrorManager.silentErrorDisable();
         
         if (!xFunction._available())
         {
            message[0] = "function '" + function + "' does not exist.";
            return false;
         }
      }
      
      // finally add/update menu as function 
      ErrorManager.silentErrorEnable();
      new FindQuery(xFunction, 
                    "rtrim(xFunction.functionName) = ?", 
                    null, 
                    "xFunction.functionName asc", 
                    new Object[] { new character(def.getName()) }).first();
      ErrorManager.silentErrorDisable();
      
      if (xFunction._available())
      {
         xFunction.setFunctionDescription(new character(def.getDescription()));
         xFunction.setMdate(date.today());
         xFunction.setType(new character("panel menu"));
      }
      else
      {
         xFunction.create();
         
         ErrorManager.silentErrorEnable();
         RecordBuffer.startBatch();

         xFunction.setFunctionName(new character(def.getName()));
         xFunction.setFunctionDescription(new character(def.getDescription()));
         xFunction.setMdate(date.today());
         xFunction.setType(new character("panel menu"));
         xFunction.setApprovalCode(new character("xt"));

         RecordBuffer.endBatch();
         ErrorManager.silentErrorDisable();
      }
      
      return true;
   }

   /**
    * Copy values from given {@link TaggedName} instance into given
    * {@link XtPrint} instance.
    * 
    * @param    tn
    *           Source of values.
    * @param    xPrint
    *           Destination of values.
    */
   private static void tagged2xPrint(final TaggedName tn,
                                     final XtPrint.Buf xPrint)
   {
      xPrint.setPrinterName(new character(tn.get(0)));
      xPrint.setPrinterType(new character(tn.get(1))); 
      xPrint.setInfile(new character(tn.get(2)));
      xPrint.setOutfile(new character(tn.get(3))); 
      xPrint.setStartStr(new character(tn.get(4))); 
      xPrint.setEndStr(new character(tn.get(5))); 
      xPrint.setStartProc(new character(tn.get(6))); 
      xPrint.setEndProc(new character(tn.get(7)));
      xPrint.setPrintSrvr(new character(tn.get(8)));
   }

   /**
    * Copy values from given {@link TaggedName} instance into given
    * {@link XtFunction} instance.
    * 
    * @param    tn
    *           Source of values.
    * @param    xFunction
    *           Destination of values.
    *           
    * @return   Parsed notes, if any.
    */
   private static character[] tagged2xFunc(final TaggedName tn,
                                           final XtFunction.Buf xFunction)
   {
      int i = 0;
      xFunction.setFunctionName(new character(tn.get(i++)));
      xFunction.setFunctionDescription(new character(tn.get(i++))); 
      xFunction.setType(new character(tn.get(i++)));
      xFunction.setFunctionId(new integer(tn.get(i++))); 
      xFunction.setKeyAccess(new character(tn.get(i++))); 
      xFunction.setPrinterName(new character(tn.get(i++))); 
      xFunction.setRunProgram(new character(tn.get(i++))); 
      xFunction.setMdate(new date(tn.get(i++)));
      xFunction.setTransactionOk(new logical(tn.get(i++)));
      xFunction.setSymixFunction(new logical(tn.get(i++)));
      xFunction.setPriority(new integer(tn.get(i++)));
      xFunction.setRunCall(new logical(tn.get(i++)));
      xFunction.setPermitFunction(new logical(tn.get(i++)));
      xFunction.setTypeDescription(new character(tn.get(i++)));
      xFunction.setApprovalCode(new character(tn.get(i++)));
      // skip over the resp dep
      i++;
      character[] notes = parseNotes(tn.get(i++));
      xFunction.setUserSecurity(new character(tn.get(i++)));
      xFunction.setOwnSecurity(new character(tn.get(i++)));
      xFunction.setUserLevel(new integer(tn.get(i++)));
      xFunction.setOwnLevel(new integer(tn.get(i++)));
      
      return notes;
   }
   
   /**
    * Get the name of the specified employee.
    * 
    * @param    empNum
    *           The employee's id.
    *           
    * @return   the employee name or <code>null</code> if doesn't exist.
    */
   public String getEmployeeName(int empNum)
   {
      return getRecordDescription("Employee", 
                                  new integer(empNum), 
                                  "employeeNumber", 
                                  "name");
   }
   
   /**
    * Get the skill-code of the specified employee.
    * 
    * @param    empNum
    *           The employee's id.
    *           
    * @return   the employee skill-code or <code>null</code> if doesn't exist.
    */
   public String getEmployeeSkillCode(int empNum)
   {
      return getRecordDescription("Employee", 
                                  new integer(empNum), 
                                  "employeeNumber", 
                                  "skillCode");
   }
   
   /**
    * Get the description of the specified skill-code .
    * 
    * @param    skillCode
    *           The skill code
    *           
    * @return   the description of the skill-code
    */
   public String getSkillCodeDescription(String skillCode)
   {
      return getRecordDescription("WorkCenter", 
                                  new character(skillCode), 
                                  "workCenter", 
                                  "description");
   }

   /**
    * Get the name of the specified customer.
    * 
    * @param    custNum
    *           The customer's id.
    *           
    * @return   the customer name or <code>null</code> if doesn't exist.
    */
   public String getCustomerName(int custNum)
   {
      return getRecordDescription("Customer", 
                                  new integer(custNum), 
                                  "customerNumber", 
                                  "name");
   }

   /**
    * Get the type of the specified printer.
    * 
    * @param    printer
    *           The printer name
    *           
    * @return   the printer type or <code>null</code> if doesn't exist.
    */
   public String getPrinterType(String printer)
   {
      if (printer == null || printer.length() == 0)
      {
         return null;
      }

      return getRecordDescription("XtPrint", 
                                  new character(printer), 
                                  "printerName", 
                                  "printerType");
   }

   /**
    * Get the description for the function specified by the entry-point.
    * 
    * @param    entryPoint
    *           The entry-point's id.
    *           
    * @return   the entry-point name or <code>null</code> if doesn't exist.
    */
   public String getEntryPointDescription(String entryPoint)
   {
      if (entryPoint == null || entryPoint.length() == 0)
      {
         return null;
      }

      return getRecordDescription("XtFunction", 
                                  new character(entryPoint), 
                                  "functionName", 
                                  "functionDescription");
   }
   
   /**
    * Get the created-date for a new account.
    * 
    * @return   the current server-side date.
    */
   public DateValue getCreatedDate()
   {
      return new DateValue();
   }
   
   /**
    * Get the next user-code value. This value still needs to be recomputed
    * on server-side if is not unique.
    * 
    * @return   next user-code value.
    */
   public String getNextUserCode()
   {
      Set<String> userCodes = getUserCodes(null);

      return getNextUserCode(userCodes);
   }
   
   /**
    * Get the next user-code value. 
    * 
    * @param    userCodes
    *           the set of existing user-code values
    * 
    * @return   next user-code value.
    */
   public static String getNextUserCode(Set<String> userCodes)
   {
      String max = "000";
      if (userCodes.size() > 0)
      {
         // determine maximum value from all codes
         Iterator<String> itr = userCodes.iterator();
         while (itr.hasNext())
         {
            String s = itr.next();
            if (max.compareTo(s) < 0)
            {
               max = s;
            }
         }
      }
      else
      {
         // this is the first account, get the first value
      	return max;
      }

      // generate an unique 3-char value, using base 36
      // (the 0123456789abcdefghijklmnopqrstuvwxyz alphabet)
      final String alph = "0123456789abcdefghijklmnopqrstuvwxyz";
      int j = alph.indexOf(max.charAt(0));
      int k = alph.indexOf(max.charAt(1));
      int l = alph.indexOf(max.charAt(2));

      boolean finished = false;
      while (!finished)
      {
        l = l + 1;
        if (l > 35)
        {
           l = 0;
           k = k + 1;
        }
        if (k > 35)
        {
           k = 0;
           j = j + 1;
        }

        if (j > 35)
        {
          finished = true;
          return null;
        }

        max = alph.substring(j, j + 1) + 
              alph.substring(k, k + 1) + 
              alph.substring(l, l + 1);

        finished = !userCodes.contains(max);
      }
      return max;
   }

   /**
    * Get the company name value from the directory.
    * 
    * @param    server
    *           the named server instance to query or "" for the default
    *           server
    * 
    * @return   the company name value or <code>null</code> if there is none
    */
   public String getCompanyName(String server)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryService();
      
      StringBuilder sb = new StringBuilder();
      sb.append("/server/");
      
      if (server.length() == 0)
      {
         sb.append("default");
      }
      else
      {
         sb.append(server);
      }

      sb.append("/majic/company");
      
      return ds.getNodeString(sb.toString(), "value");
   }
   
   /**
    * Set the company name value in the directory.
    * 
    * @param    server
    *           the named server instance to change or "" for the default
    *           server; the instance of majic/company must exist there
    * 
    * @param    companyName
    *           the new value for the company name
    * 
    * @return   <code>true</code> if set successfully
    */
   public boolean setCompanyName(String server, String companyName)
   {
      DirectoryService ds = AdminServerImpl.getDirectoryService();
      
      StringBuilder sb = new StringBuilder();
      sb.append("/server/");
      
      if (server.length() == 0)
      {
         sb.append("default");
      }
      else
      {
         sb.append(server);
      }

      sb.append("/majic/company");
      
      // open editing batch for the branch
      String branch = sb.toString();
      
      boolean rc = ds.openBatch(branch);
      if (!rc)
      {
         message("Can't lock directory branch. Try later."); 
         return false;
      };
      
      // change the value
      rc = ds.setNodeString(branch, "value", companyName);
      
      // check directory operations
      if (!rc)
      {
         message("Directory modification failed for " + branch);
      }
      
      // close batch with commit or rollback and no refresh
      boolean rcc = ds.closeBatch(rc, false); 
      if (!rcc)
      {
         message("Error unlocking directory branch.");
      }
      return rc;
   }
   
   /**
    * Generate a filter for the given column and value. Depending on value's
    * type, it will create a character filter using <code>upper(rtrim(s.&lt;col&gt;, ' \t\n\r')) like ?</code>
    * or a filter like <code>s.&lt;col&gt; >= ?</code> or
    * <code>s.&lt;col&gt; = ?</code> or <code>s.&lt;col&gt; <= ?</code>,
    * depending on the <code>mode</code> parameter.
    * 
    * @param    col
    *           The column name.
    * @param    buf
    *           HQL expression to which the filter expression will be added.
    * @param    mode
    *           Filtering mode.
    */
   private void addFilter(String col,
                          BaseDataType filter,
                          HQLExpression buf,
                          int mode)
   {
      if (filter instanceof character)
      {
         buf.append("upper(rtrim(s.");
         buf.append(col);
         buf.append(", ' \t\n\r')) like ", true);
         character c = (character) filter;
         c.assign(TextOps.toUpperCase(c));
      }
      else
      {
         buf.append("s.");
         buf.append(col);
         
         switch(mode)
         {
            case MODE_GE:
               buf.append(" >= ", true);
               break;
            case MODE_LE:
               buf.append(" <= ", true);
               break;
            case MODE_EQ:
               buf.append(" = ", true);
               break;
         }
      }
   }

   /**
    * Get count of ACL's, per each function.
    * 
    * @return   a map of ACL's.
    */
   public Map<String, Integer> getACLs()
   {
      Map<String, Integer> acls = new HashMap<String, Integer>();
      
      String[] res = AdminServerImpl.listAclInstances("majic-function");
      String name = AdminServerImpl.getServerName();
      
      if (name == null)
         name = "";
      
      for (String instance : res)
      {
         if (instance.length() == 0 || instance.equals(name))
         {
            AclDef def = AdminServerImpl.getAcl("majic-function", instance);
            
            for (AclDef.Acl oneAcl: def.acls)
            {
               Integer count = acls.get(oneAcl.name);

               if (count == null)
                  acls.put(oneAcl.name, Integer.valueOf(1));
               else
                  acls.put(oneAcl.name, Integer.valueOf(count.intValue() + 1));
            }
         }
      }

      return acls;
   }
   
   /**
    * Get value of the specified column of the specified entity, with the
    * specified id.
    * 
    * @param    entity
    *           The entity class-name from <code>majic</code> database.
    * @param    id
    *           The id of the entity
    * @param    idCol
    *           The name of the column which holds entity's id
    * @param    column
    *           The name of the column which holds the required value
    * 
    * @return   the value of the specified column of the entity with the given
    *           id, or <code>null</code> if no record found
    */
   private BaseDataType getRecordField(final String entity,
                                       final BaseDataType id,
                                       final String idCol,
                                       final String column)
   {
      Object rc = null;
      
      try
      {
         rc = StandardServer.invoke(2, new Isolatable()
         {
            public Object execute()
            {
               return getRecordFieldWorker(entity, id, idCol, column);
            }
         });
      }
      catch (IllegalAccessException iax)
      {
         // never the case here
      }
      catch (InvocationTargetException itx)
      {
         // never the case here
      }
      catch (java.lang.InstantiationException ix)
      {
         // never the case here
      }
    
      return (BaseDataType)rc;
   }

   /**
    * Get the description for the specified entity, with the specified id.
    *
    * @param    entity
    *           The entity class-name from <code>majic</code> database.
    * @param    id
    *           The id of the entity
    * @param    idCol
    *           The name of the column which holds entity's id
    * @param    fieldCol
    *           The name of the column which holds entity's description
    *
    * @return   the description for the entity with the give id, or
    *                     <code>null</code> if no record found
    */
   private String getRecordDescription(final String entity,
                                       final BaseDataType id,
                                       final String idCol,
                                       final String descCol)
   {
      BaseDataType res = getRecordField(entity, id, idCol, descCol);
      return res == null ? null : TaggedName.prettyPrint(res);
   }
   
   /**
    * Get the records in the specified entity, in the specified page.
    * <p>
    * If any filters are provided, they will be applied.
    * 
    * @param    entity
    *           The entity class-name from <code>majic</code> database.
    * @param    idFilter
    *           The filter to be applied to the id column
    * @param    nameFilter
    *           The filter to be applied to the description column
    * @param    page
    *           The current page
    * @param    pageSize
    *           The page size.
    * @param    idColPos
    *           Position of ID column in the resulting output. 
    * @param    mode
    *           Filtering mode. See <code>MODE_*</code> constants.
    * @param    idCol
    *           The name of the column which holds entity's id
    * @param    cols
    *           Remaining names of columns to retrieve
    * @return   a record list, or <code>null</code> if none found.
    */
   private TaggedName[] listRecords(final String entity,
                                    final BaseDataType idFilter, 
                                    final BaseDataType nameFilter, 
                                    final int page, 
                                    final int pageSize, 
                                    final int idColPos, 
                                    final int mode, 
                                    final String idCol, 
                                    final String ... cols)
   {
      Object rc = null;
      
      try
      {
         rc = StandardServer.invoke(2, new Isolatable()
         {
            public Object execute()
            {
               return listRecordsWorker(entity, idFilter, nameFilter, page,
                                        pageSize, idColPos, mode, idCol, cols);
            }
         });
      }
      catch (IllegalAccessException iax)
      {
         // never the case here
      }
      catch (InvocationTargetException itx)
      {
         // never the case here
      }
      catch (java.lang.InstantiationException ix)
      {
         // never the case here
      }
    
      return (TaggedName[])rc;
   }
   
   /**
    * Get the records in the specified entity, in the specified page.
    * <p>
    * If any filters are provided, they will be applied.
    * 
    * @param    entity
    *           The entity class-name from <code>majic</code> database.
    * @param    idFilter
    *           The filter to be applied to the id column
    * @param    nameFilter
    *           The filter to be applied to the description column
    * @param    page
    *           The current page
    * @param    pageSize
    *           The page size.
    * @param    idColPos
    *           Position if ID column in resulting set.
    * @param    mode
    *           Filtering mode.
    * @param    idCol
    *           The name of the column which holds entity's id
    * @param    cols
    *           Array of names of other columns to retrieve. Zero arrray
    *           element represents the name of the column to be used in the
    *           name filter.
    * @return   a record list, or <code>null</code> if none found.
    */
   @SuppressWarnings("null")
   private TaggedName[] listRecordsWorker(String entity,
                                          BaseDataType idFilter, 
                                          BaseDataType nameFilter, 
                                          int page, 
                                          int pageSize, 
                                          int idColPos, 
                                          int mode, 
                                          String idCol, 
                                          String ... cols)
   {
      Persistence p = null;
      
      try
      {
         p = PersistenceFactory.getInstance("majic");
      }
      catch (Exception e) 
      {
         message(e.getMessage());
         message("error connecting to majic database");
         return null;
      }
      
      HQLExpression hql = new HQLExpression("select ");
      
      if (cols  != null && cols.length > 0)
      {
         int cnt = 0;
         
         for (String col : cols)
         {
            if (cnt == idColPos)
            {
               if (cnt > 0)
                  hql.append(", ");
               addColumn(hql, idCol);
               cnt++;
            }

            if (cnt > 0)
               hql.append(", ");
            addColumn(hql, col);

            cnt++;
         }
      }
      
      hql.append(" from aero.timco.majic.dmo.majic.");
      hql.append(entity);
      hql.append(" s ");

      Object[] values = null;
      Type[] types = null;
      
      boolean hasId = idFilter != null && !idFilter.isUnknown();
      boolean hasName = nameFilter != null 
                        && !nameFilter.isUnknown() 
                        && cols != null 
                        && cols.length > 0;
      
      if (hasId || hasName)
      {
         hql.append(" where ");
         
         if (hasId)
         {
            addFilter(idCol, idFilter, hql, mode);

            if (!hasName)
            {
               values = new Object[] { idFilter };
            }
         }

         if (hasId && hasName)
         {
            hql.append(" and ");
            values = new Object[] { idFilter, nameFilter };
         }

         if (hasName)
         {
            // name filter is not strict
            addFilter(cols[0], nameFilter, hql, mode);

            if (!hasId)
            {
               values = new Object[] { nameFilter };
            }
         }
         
         types = DBUtils.makeTypeArray(values);
      }
      boolean upper = idFilter instanceof character;
      
      hql.append(" order by ");
      hql.append((upper ? "upper(rtrim(" : ""));
      hql.append("s.");
      hql.append(idCol); 
      hql.append(upper ? ", ' \t\n\r'))" : "");

      TaggedName[] tn = null;

      try
      {
         int first = (pageSize > 0) ? page * pageSize : 0;

         ScrollableResults s = p.scroll(hql.toFinalExpression(),
                                        values, 
                                        types, 
                                        pageSize, 
                                        first);

         // if some records were found
         if (s.first())
         {            
            if (pageSize > 0)
            {
               // get the next pageSize records
               tn = new TaggedName[pageSize];

               boolean last = false;

               for (int i = 0; i < pageSize && !last; i++)
               {
                  tn[i] = new TaggedName(s.get());

                  last = !s.next();
               }
            }
            else
            {
               ArrayList<TaggedName> list = new ArrayList<TaggedName>();
               
               do
               {
                  list.add(new TaggedName(s.get()));
               }
               while (s.next());
               
               tn = list.toArray(new TaggedName[list.size()]);
            }
         }
      }
      catch (Exception e) 
      {
         String f1 = TaggedName.prettyPrint(idFilter);
         String f2 = TaggedName.prettyPrint(nameFilter);
         
         message("error retrieving " + entity + " records with filters " + 
                  "<" + f1 + "> and <" + f2 + ">");
      }
      finally
      {
         // remove the state synchronizer
         Session session = SessionManager.get().getSession();
         session.registerSynchronizer(null);
      }
      
      return tn;      
   }

   /**
    * Add column expression to SELECT clause with proper separators.
    * 
    * @param    hql
    *           Work buffer.
    * @param    col
    *           Column name.
    */
   private static void addColumn(HQLExpression hql, String col)
   {
      hql.append((Character.isLetter(col.charAt(0)) ? "s." : ""));
      hql.append(col);
   }   
   
   /**
    * Get the value of the specified column of the specified entity, with the
    * specified id.
    * 
    * @param    entity
    *           The entity class-name from <code>majic</code> database.
    * @param    id
    *           The id of the entity
    * @param    idCol
    *           The name of the column which holds entity's id
    * @param    column
    *           The name of the column which holds the required value
    * 
    * @return   the value of the specified column of the entity with the given
    *           id, or <code>null</code> if no record found
    */
   @SuppressWarnings("unchecked")
   private BaseDataType getRecordFieldWorker(String entity,
                                             BaseDataType id,
                                             String idCol,
                                             String column)
   {
      Persistence p = null;
      
      try
      {
         p = PersistenceFactory.getInstance("majic");
      }
      catch (Exception e) 
      {
         message(e.getMessage());
         message("error connecting to majic database");
         return null;
      }
      
      boolean upper = (id instanceof character);
      
      StringBuilder buf = new StringBuilder("select s.");
      buf.append(column);
      buf.append(" from aero.timco.majic.dmo.majic.");
      buf.append(entity);
      buf.append(" s where ");
      if (upper)
      {
         buf.append("upper(rtrim(");
      }
      buf.append("s.");
      buf.append(idCol);
      if (upper)
      {
         buf.append(", ' \t\n\r'))");
      }
      buf.append(" = ?0");
      String hql = buf.toString();
      
      if (upper)
      {
         character c = (character) id;
         c.assign(TextOps.toUpperCase(c));
      }
      Object[] values = new Object[] { id };
      Type[] types = DBUtils.makeTypeArray(values);
      
      BaseDataType desc = null;
      try
      {
         List<BaseDataType> list = p.list(hql, values, types, 1, 0, true);
         if (list != null)
         {
            desc = list.get(0);
         }
      } 
      catch (Exception e)
      {
         message("error retrieving " + entity + " record " + TaggedName.prettyPrint(id));
      }
      finally
      {
         // remove the state synchronizer
         Session session = SessionManager.get().getSession();
         session.registerSynchronizer(null);
      }

      return desc;
   }
   
   /**
    * Get the "User Authorization Form" report with the given parameters.
    * 
    * @param  startingAccount
    *         The starting account. May be <code>null</code>.
    * @param  endingAccount
    *         The starting account. May be <code>null</code>.
    * @param  startingACL
    *         The starting function. May be <code>null</code>.
    * @param  endingACL
    *         The ending function. May be <code>null</code>.
    *         
    * @return an array of {@link TaggedName} instances, with the report rows.
    *         When data for a new user starts, the first row for this user
    *         is the header data.
    */
   public TaggedName[] getReportUserAuthForm(String startingAccount, 
                                             String endingAccount,
                                             String startingACL,
                                             String endingACL)
   {
      if (startingAccount != null && startingAccount.length() == 0)
         startingAccount = null;
      if (endingAccount != null && endingAccount.length() == 0)
         endingAccount = null;
      if (startingACL != null && startingACL.length() == 0)
         startingACL = null;
      if (endingACL != null && endingACL.length() == 0)
         endingACL = null;
      
      List<TaggedName> res = new ArrayList<TaggedName>();
      
      TaggedName[] groupsTN = AdminServerImpl.listGroups();
      TaggedName[] usersTN  = AdminServerImpl.listUsers();
      TaggedName[] procsTN  = AdminServerImpl.listProcesses();
      
      Set<String> accounts = new TreeSet<String>();
      TaggedName[][] allAccs = { usersTN, groupsTN, procsTN };
      
      for (int t = 0; t < allAccs.length; t++)
      {
         TaggedName[] accs = allAccs[t];
         
         for (int i = 0; i < accs.length; i++)
         {
            String acc = accs[i].getName().toLowerCase();
            
            if ((startingAccount == null || 
                 acc.compareTo(startingAccount) >= 0) &&
                (endingAccount == null   || 
                 acc.compareTo(endingAccount) <= 0))
            {
               accounts.add(acc);
            }
         }
      }
      
      // add all groups to the list
      Set<String> groups = new TreeSet<String>();
      for (int i = 0; i < groupsTN.length; i++)
      {
         String acc = groupsTN[i].getName().toLowerCase();

         groups.add(acc);
      }
      
      // get the private and shared ACLs for the majic-function resource
      AclDef aShar = AdminServerImpl.getAcl("majic-function", "");
      AclDef aPriv = AdminServerImpl.getAcl("majic-function", 
                                            AdminServerImpl.getServerName());

      AclDef[] functionAcls = { aShar, aPriv };
      
      // save the responsible departments for each function
      Map<String, String> respDepts = new HashMap<String, String>();

      // a cache of ACLs for groups to which the user accounts are assigned
      Map<String, Map<String, List<Rights>>> groupAcls = 
         new HashMap<String, Map<String,List<Rights>>>();

      // the ACLs which will be output by the report
      Map<String, Map<String, List<Rights>>> acls = 
         new HashMap<String, Map<String, List<Rights>>>();

      for (int t = 0; t < functionAcls.length; t++)
      {
         AclDef functionACLs = functionAcls[t];
         Map<Integer, String> subjects = functionACLs.subjects;
      
         for (int i = 0; i < functionACLs.acls.length; i++)
         {
            AclDef.Acl functionACL = functionACLs.acls[i];
            String function = functionACL.name;
   
            if ((startingACL == null || 
                 function.compareTo(startingACL) >= 0) &&
                (endingACL == null   || 
                 function.compareTo(endingACL) <= 0))
            {
               if (!respDepts.containsKey(function))
               {
                  respDepts.put(function, Config.getDeptForFunction(function));
               }
               
               // for all its subjects
               for (int subjIdx : subjects.keySet())
               {
                  String subject = subjects.get(subjIdx);
                  Map<String, List<Rights>> accACL = null;
                  
                  if (accounts.contains(subject))
                  {
                     // if we are interested about this subject
                     accACL = acls.get(subject);
                     if (accACL == null)
                     {
                        accACL = new HashMap<String, List<Rights>>();
                        acls.put(subject, accACL);
                     }
                  }

                  if (accACL == null && groups.contains(subject))
                  {
                     // if this is a group, determine its ACL's, just in case
                     // we need to resolve a user's ACLs who belongs to this
                     // group
                     accACL = groupAcls.get(subject);
                     if (accACL == null)
                     {
                        accACL = new HashMap<String, List<Rights>>();
                        groupAcls.put(subject, accACL);
                     }
                  }
                  
                  if (accACL != null)
                  {
                     // then add all ACL rights where current subject appears  
                     // to its info
                       
                     for (int k = 0; k < functionACL.subjects.length; k++)
                     {
                        if (functionACL.subjects[k] == subjIdx)
                        {
                           List<Rights> rights = accACL.get(function);
                           if (rights == null)
                           {
                              rights = new ArrayList<Rights>();
                              accACL.put(function, rights);
                           }
                           
                           rights.add(functionACL.rights);
                        }
                     }
                  }                
               }
            }
         }
      }
      
      for (String acc : accounts)
      {
         // get the ACLs
         Map<String, List<Rights>> accAcl = acls.get(acc);
         if (accAcl == null)
            accAcl = new HashMap<String, List<Rights>>();

         // for user accounts, the group ACL's must be taken too
         Map<String, List<Rights>> groupAcl = 
            new HashMap<String, List<Rights>>();

         String description = null;
         String date = null;
         
         {
            UserDef user = AdminServerImpl.getUser(acc);
            if (user != null)
            {
               description = user.person;
               date = user.date == null ? "" : user.date.toString();
               
               for (int i = 0; i < user.groups.length; i++)
               {
                  String group = user.groups[i];
                  
                  Map<String, List<Rights>> gAcl = acls.get(group);
                  
                  if (gAcl == null)
                  {
                     // if not among the reported ones, get it from the global
                     // cache
                     gAcl = groupAcls.get(group);
                  }
                  
                  if (gAcl != null)
                  {
                     // if the group was found, collect its ACLs
                     groupAcl.putAll(gAcl);
                  }
               }
            }
            else
            {
               ProcessDef proc = AdminServerImpl.getProcess(acc);
               
               if (proc != null)
                  description = proc.description;
               else
               {
                  GroupDef grpDef = AdminServerImpl.getGroup(acc);
                  if (grpDef != null)
                     description = grpDef.description;
               }
            }
         }

         if (accAcl.isEmpty() && groupAcl.isEmpty())
            continue;

         MajicExtDef ext = getAccountExtensions(acc);
         String[] values = new String[8];

         if (ext == null)
         {
            // get the header
            values[0] = acc;
            values[1] = "";
            values[2] = description == null ? "" : description;
            values[3] = "        -";
            values[4] = "";
            values[5] = "";
            values[6] = "";
            values[7] = "        -";
         }
         else
         {
            String empName = getEmployeeName(ext.empNum);
            String skillCode = empName == null  
                                  ? null : getEmployeeSkillCode(ext.empNum);
            String skillName = skillCode == null 
                                  ? null : getSkillCodeDescription(skillCode);

            // get the header
            values[0] = acc;
            values[1] = ext.userNum == 0 ? "" 
                                       : new integer(ext.userNum)
                                         .toString(">>>>>>>");
            values[2] = description == null ? "" : description;
            values[3] = empName == null 
                        ? "        -"
                        : new integer(ext.empNum).toString(">>>>>>>") + 
                          " - " + empName;
            values[4] = ext.createdDate == null ? "" 
                                                : ext.createdDate.toString();
            values[5] = date == null ? "" : date;
            values[6] = ext.entryPoint == null ? "" : ext.entryPoint;
            values[7] = skillName == null 
                           ? "        -"
                           : new character(skillCode).toString("x(7)") + 
                             " - " + skillName;
            
         }
         res.add(new TaggedName(values));
         
         Set<String> funs = new TreeSet<String>(accAcl.keySet());
         // remove all the user's ACLs from the group's map
         for (String fun : funs)
         {
            if (groupAcl.isEmpty())
            {
               break;
            }

            groupAcl.remove(fun);
         }
         
         // merge the group ACLs with the user's ACLs
         accAcl.putAll(groupAcl);
         
         funs = new TreeSet<String>(accAcl.keySet());
         // get data for each acl defined at account or group level
         for (String fun : funs)
         {
            List<Rights> rights = accAcl.get(fun);
            
            for (Rights right : rights)
            {
               FunctionRights fr = (FunctionRights) right;
               String depts = respDepts.get(fun);
                
               String[] aclVals = new String[9];
               aclVals[0] = acc;
               aclVals[1] = fun;
               aclVals[2] = fr.get(FunctionRights.BIT_RUN_PERMISSION) 
                               ? "Yes" : "No";
               aclVals[3] = fr.get(FunctionRights.BIT_ADD_PERMISSION) 
                               ? "Yes" : "No";
               aclVals[4] = fr.get(FunctionRights.BIT_UPDATE_PERMISSION) 
                               ? "Yes" : "No";
               aclVals[5] = fr.get(FunctionRights.BIT_DELETE_PERMISSION) 
                               ? "Yes" : "No";
               aclVals[6] = fr.get(FunctionRights.BIT_COSTS_PERMISSION) 
                               ? "Yes" : "No";
               aclVals[7] = fr.get(FunctionRights.BIT_PRICES_PERMISSION) 
                               ? "Yes" : "No";
               aclVals[8] = depts == null ? "" : depts;
               
               res.add(new TaggedName(aclVals));
            }
         }
      }
      
      return res.toArray(new TaggedName[0]);
   }
   
   /**
    * Get the "User Authorization By Function" report with the given
    * parameters.
    *
    * @param startingACL
    *        Starting ACL.
    * @param endingACL
    *        Ending ACL.
    *
    * @return an array of {@link TaggedName} instances, with the report rows.
    */
   public TaggedName[] getReportUserAuthByFunction(String startingACL, 
                                                   String endingACL)
   {
      if (startingACL != null && startingACL.length() == 0)
         startingACL = null;
      if (endingACL != null && endingACL.length() == 0)
         endingACL = null;

      List<TaggedName> res = new ArrayList<TaggedName>();
      Map<Integer, String> empNames = new HashMap<Integer, String>();
      Map<String, MajicExtDef> extensions = new HashMap<String, MajicExtDef>();
      Map<String, String> respDepts = new HashMap<String, String>();
      Map<String, String> terminations = new HashMap<String, String>();
      
      // get the private and shared ACLs for the majic-function resource
      AclDef aShar = AdminServerImpl.getAcl("majic-function", "");
      AclDef aPriv = AdminServerImpl.getAcl("majic-function", 
                                            AdminServerImpl.getServerName());
      
      AclDef[] allAcls = { aShar, aPriv };
      for (int t = 0; t < allAcls.length; t++)
      {
         AclDef functionACLs = allAcls[t];
         Map<Integer, String> subjects = functionACLs.subjects;

         for (int i = 0; i < functionACLs.acls.length; i++)
         {
            AclDef.Acl functionACL = functionACLs.acls[i];
            String function = functionACL.name;
   
            if ((startingACL == null || 
                 function.compareTo(startingACL) >= 0) &&
                (endingACL == null   || 
                 function.compareTo(endingACL) <= 0))
              {
                 String dept = null;
                 if (respDepts.containsKey(function))
                 {
                    dept = respDepts.get(function);
                 }
                 else
                 {
                    dept = Config.getDeptForFunction(function);
                    respDepts.put(function, dept);
                 }

                 FunctionRights fr = (FunctionRights) functionACL.rights;
                 
                 // for all its subjects
                 for (int k = 0; k < functionACL.subjects.length; k++)
                 {
                    String account = subjects.get(functionACL.subjects[k]);
                    
                    // get the extensions
                    MajicExtDef ext = null;
                    if (extensions.containsKey(account))
                    {
                       ext = extensions.get(account);
                    }
                    else
                    {
                       ext = getAccountExtensions(account);
                       extensions.put(account, ext);
                    }
                    
                    String empNum = "0";
                    String empName = "";
                    
                    if (ext != null)
                    {
                       empNum = Integer.toString(ext.empNum);
                       if (ext.empNum != 0)
                       {
                          if (empNames.containsKey(ext.empNum))
                          {
                             empName = empNames.get(ext.empNum);
                          }
                          else
                          {
                             empName = getEmployeeName(ext.empNum);
                             if (empName == null)
                                empName = "";

                             empNames.put(ext.empNum, empName);
                          }
                       }
                    }
                    
                    String termination = "";
                    if (terminations.containsKey(account))
                    {
                       termination = terminations.get(account);
                    }
                    else
                    {
                       if (ext != null && ext.empNum > 0)
                       {
                          date termDate = 
                             (date) getRecordField("Employee",
                                                   new integer(ext.empNum),
                                                   "employeeNumber",
                                                   "terminationDate");
                          termination = 
                             TaggedName.prettyPrintNoUnknows(termDate);
                       }
                       
                       terminations.put(account, termination);
                    }

                    String[] cols = new String[12];
                    cols[0] = function;
                    cols[1] = empName;
                    cols[2] = empNum;
                    cols[3] = termination;
                    cols[4] = fr.get(FunctionRights.BIT_RUN_PERMISSION) 
                                 ? "Yes" : "No";
                    cols[5] = fr.get(FunctionRights.BIT_ADD_PERMISSION) 
                                 ? "Yes" : "No";
                    cols[6] = fr.get(FunctionRights.BIT_UPDATE_PERMISSION) 
                                 ? "Yes" : "No";
                    cols[7] = fr.get(FunctionRights.BIT_DELETE_PERMISSION) 
                                 ? "Yes" : "No";
                    cols[8] = fr.get(FunctionRights.BIT_COSTS_PERMISSION) 
                                 ? "Yes" : "No";
                    cols[9] = fr.get(FunctionRights.BIT_PRICES_PERMISSION) 
                                 ? "Yes" : "No";
                    cols[10] = account;
                    cols[11] = dept;
                    
                    res.add(new TaggedName(cols));
                 }
              }
         }
      }
      
      // sort the results by function name and user name
      Collections.sort(res, new Comparator<TaggedName>()
      {
         public int compare(TaggedName tn1, TaggedName tn2)
         {
            String f1 = tn1.get(0);
            String f2 = tn2.get(0);
            
            if (f1.equals(f2))
            {
               String u1 = tn1.get(10);
               String u2 = tn2.get(10);
               
               return u1.compareTo(u2);
            }
            
            return f1.compareTo(f2);
         }
      });
      
      return res.toArray(new TaggedName[0]);
   }


   /**
    * Utility to generate new user-num, user-code, key and created-date 
    * extension fields for a specified account. This is done only if the
    * account is missing that field.
    * <p>
    * The syntax is:
    * 
    * java aero.timco.majic.admin.server.MajicAccounts &lt;directory.xml&gt; 
    * &lt;account name&gt;
    *  
    * @param    args
    *           Application command line parameters.
    *           
    * @throws   ConfigurationException
    *           In case of configuration error.
    */
   public static void main(String[] args)
   throws ConfigurationException
   {
      boolean syntax = false;
      String account = null;

      if (args.length == 3)
         syntax = !args[0].endsWith(".xml") || 
                  args[1].endsWith(".xml") ||
                  args[2].endsWith(".xml");
      else
         syntax = true;
      
      if (syntax)
      {
         System.out.println("usage: java MajicAccounts "
            + "[<directory.xml>] account");
         return;
      }

      // Empty BootstrapConfig
      BootstrapConfig bc = new BootstrapConfig(null, null, null, null);
      
      // directory package configuration
      bc.setServer(true);
      bc.setConfigItem("directory", "schema",  "source",   "schema.xml");
      bc.setConfigItem("directory", "backend", "type",     "xml");
      bc.setConfigItem("directory", "xml",     "filename", args[0]);
      bc.setConfigItem("security",  "server",  "id",       args[1]);

      account = args[2];

      System.out.println("Instantiating DirectoryService");
      DirectoryService dir = DirectoryService.createInstance(bc);
      System.out.println("DirectoryService is up");

      System.out.println("Binding to directory.");
      if (!dir.bind())
         throw new RuntimeException("bind() failed");
      System.out.println("Bound to directory.");

      String path = "/security/accounts/users/" + account;
      
      // opening the account 
      System.out.println("Testing user editing batch...");
      if (dir.openBatch(path))
         System.out.println("Opened an editing batch for " +
         		             "/security/accounts/users/" + account);
      else
      {
         path = "/security/accounts/processes/" + account;

         System.out.println("Testing process editing batch...");
         if (!dir.openBatch(path))
            throw new RuntimeException("openBatch() failed");
         System.out.println("Opened an editing batch for " +
               "/security/accounts/processes/" + account);
      }
      
      Connection conn = connectToDatabase(dir);
      
      // update the user-num, user-code, key and created-date extensions
      addAccountExtensions(dir, conn, path);
      
      try
      {
         System.out.println("Closing DB connection...");
         conn.close();
         System.out.println("DB connection closed.");
      } 
      catch (SQLException e)
      {
         e.printStackTrace();
         System.out.println("Error closing DB connection.");
      }
      
      System.out.println("Closing the batch...");
      if (!dir.closeBatch(true))
         throw new RuntimeException("closeBatch() failed");
      System.out.println("Batch closed.");
      System.out.println("Unbinding from directory...");
      if (!dir.unbind())
         throw new RuntimeException("unbind() failed");
      System.out.println("Unbound.");
 
      System.out.println("Done.");
   }
   
   /**
    * Add user-num, user-code, key and created-date extensions for the 
    * specified account. This is done only for the extensions which are not
    * defined for the specified account.
    * 
    * @param   dir
    *          The directory service.
    * @param   conn
    *          A JDBC connection to the Majic database.
    * @param   path
    *          The account path in the directory.
    */
   private static void addAccountExtensions(DirectoryService dir,
                                            Connection conn,
                                            String path)
   {
      // check the user-code
      String node = path + "/user-code";
      String oldCode = dir.getNodeString(node, "value");
      if (oldCode != null)
      {
         System.out.println(node + " already exists with value = " + oldCode);
      }
      else
      {
         Set<String> userCodes = listUserCodes(dir);
         String userCode = getNextUserCode(userCodes);
         
         if (userCode == null)
         {
            System.out.println("reached limit of user-code values (zzz)");
         }
         else
         {
            addExtension(dir, node, "string", userCode);
         }
      }
      
      // check the user-num
      node = path + "/user-num";
      Integer oldNum = dir.getNodeInteger(node, "value");
      if (oldNum != null)
      {
         System.out.println(node + " already exists with value = " + oldNum);
      }
      else
      {
         try
         {
            Integer userNum = createXtNotes(conn, "");

            addExtension(dir, node, "integer", userNum);
         } 
         catch (AccountExtValidationException e)
         {
            System.out.println("could not generate user-num entry in " +
                               "xt_notes; " + e.getMessage());
         }
      }

      // check the key
      node = path + "/key";
      Integer oldKey = dir.getNodeInteger(node, "value");
      if (oldKey != null)
      {
         System.out.println(node + " already exists with value = " + oldKey);
      }
      else
      {
         try
         {
            Integer userKey = createXtNotes(conn, 
               dir.getNodeString(path, "description"));

            addExtension(dir, node, "integer", userKey);
         } 
         catch (AccountExtValidationException e)
         {
            System.out.println("could not generate key entry in " +
                               "xt_notes; " + e.getMessage());
         }
      }

      // set the created date
      node = path + "/created-date";
      DateValue oldDate = dir.getNodeDate(node, "value");
      if (oldDate != null)
      {
         System.out.println(node + " already exists with value = " + oldDate);
      }
      else
      {
         DateValue userDate = new DateValue();

         addExtension(dir, node, "date", userDate);
      }
   }
   
   /**
    * Get all the user-codes from all the accounts in the directory.
    * 
    * @param   dir
    *          The directory service.
    *          
    * @return  A set of user-codes for all existing accounts.
    */
   private static Set<String> listUserCodes(DirectoryService dir)
   {
      Set<String> userCodes = new HashSet<String>();
      final String[] paths = {
                              "/security/accounts/processes/", 
                              "/security/accounts/users/"
                             };
      
      for (int i = 0; i < paths.length; i++)
      {
         String[] accounts = dir.enumerateNodes(paths[i]);
         
         for (int j = 0; j < accounts.length; j++)
         {
            String path = paths[i] + accounts[j] + "/user-code";
            String code = dir.getNodeString(path, "value");
            if (code != null && code.length() > 0)
            {
               userCodes.add(code);
            }
         }
      }
      
      return userCodes;
   }
   
   /**
    * Insert a record in the xt_notes, using the specified JDBC connection
    * and text.
    *  
    * @param   conn
    *          A JDBC connection to the Majic table.
    * @param   text
    *          The required text.
    *          
    * @return  The xt_notes.key value for the created record.
    */
   private static Integer createXtNotes(Connection conn, String text) 
   {
      Integer key = null;
      final String insert = "insert into xt_notes (id, key, seq, text) " +
      		                "values ((select max(id) from xt_notes) + 1, " +
      		                     "(select max(key) from xt_notes) + 1, ?, ?)";
      final String sql = "select max(key) from xt_notes";
      ResultSet rs = null;
      PreparedStatement stmt = null;
      
      try
      {
         stmt = conn.prepareStatement(insert);
         stmt.setObject(1, 1);
         stmt.setObject(2, text);
         stmt.execute();
         
         stmt = conn.prepareStatement(sql);
         rs = stmt.executeQuery();
         if (rs.next())
         {
            key = rs.getInt(1);
         }
      }
      catch (SQLException e) 
      {
         key = null;
         e.printStackTrace();
      }
      finally
      {
         if (rs != null)
         {
            try
            {
               rs.close();
            } 
            catch (SQLException e)
            {
               // ignore
            }
         }
      }
      
      return key;
   }
   
   /**
    * Add an extension to the specified account in the directory.
    * 
    * @param   dir
    *          The directory service.
    * @param   node
    *          The account node. 
    * @param   clazz
    *          The class of the extension field.
    * @param   value
    *          The value of the extension field.
    */
   private static void addExtension(DirectoryService dir,
                                    String node, 
                                    String clazz, 
                                    Object value)
   {
      Attribute attr = new Attribute(
            dir.getClassNodeAttribute(clazz, "value"),
            new Object[] { value }
            );
      
      if (value != null && dir.addNode(node, clazz, new Attribute[] { attr }))
      {
         System.out.println(node + " set to value = " + value);
      }
      else
      {
         System.out.println("Error setting " + node + " to value = " + value);
      }
   }
   
   /**
    * Obtain a JDBC connection to the permanent database as configured in the  
    * specified directory.
    * 
    * @param   dir
    *          The directory service.
    *           
    * @return  A JDBC connection to the permanent database.
    */
   private static Connection connectToDatabase(DirectoryService dir)
   {
      System.out.println("Connecting to DB...");

      // try to connect to the perm database
      String dbID = Utils.findDirectoryNodePath(dir,
                                                "database", 
                                                "container",
                                                false);

      // Read the names of all permanently connected databases.
      String[] dbNodes = (dbID == null ? null : dir.enumerateNodes(dbID));
      String dbName = null;
      if (dbNodes == null || dbNodes.length == 0)
      {
         throw new RuntimeException("No databases are configured !");
      }
      else
      {
         // find the non-temp database
         for (int i = 0; i < dbNodes.length; i++)
         {
            if (!dbNodes.equals("_temp"))
            {
               dbName = dbNodes[i];
               break;
            }
         }
         
         if (dbName == null)
         {
            throw new RuntimeException("No perm dbs are configured !");
         }
      }
      
      // connect via JDBC
      Connection conn = null;
      String hibernatePath = dbID + "/" + dbName + "/hibernate/connection/";
      String url = dir.getNodeString(hibernatePath + "/url", "value");
      String driver = dir.getNodeString(hibernatePath + "/driver_class", 
                                        "value");
      String userName = dir.getNodeString(hibernatePath + "/username", 
                                          "value");
      String password = dir.getNodeString(hibernatePath + "/password", 
                                          "value");
      try
      {
         Class.forName(driver).newInstance();
         conn = DriverManager.getConnection(url, userName, password);
      } 
      catch (Exception e)
      {
         throw new RuntimeException(e);
      }
      
      System.out.println("Connected to " + url);
      return conn;
   }

   /**
    * Determine whether the given string value is contained into the given
    * text range.
    *
    * @param  start
    *         Lower limit of the range.
    * @param  end
    *         Upper limit of the range.
    * @param  value
    *         Tested value.
    *
    * @return <code>true</code> if the value is contained into the given
    *         range.
    */
   private boolean intoStringRange(String start, String end, String value)
   {
      if (value == null)
      {
         return start.isEmpty();
      }

      if ((!start.isEmpty() && start.compareTo(value) > 0) ||
          (!end.isEmpty()   && value.compareTo(end) > 0))
      {
         return false;
      }

      return true;
   }

   /**
    * Determine whether the given integer value is contained into the given
    * integer range.
    *
    * @param  start
    *         Lower limit of the range.
    * @param  end
    *         Upper limit of the range.
    * @param  value
    *         Tested value.
    *
    * @return <code>true</code> if the value is contained into the given
    *         range.
    */
   private boolean intoIntRange(int start, int end, int value)
   {
      // [0, 0] is an empty range
      if (start == 0 && end == 0)
      {
         return true;
      }

      if (start > value ||
          value > end)
      {
         return false;
      }

      return true;
   }
}
